from collections.abc import Sequence

import stim
import svg

from tqec.circuit.qubit import GridQubit
from tqec.compile.observables.builder import Observable
from tqec.utils.frozendefaultdict import FrozenDefaultDict
from tqec.visualisation.computation.errors import get_errors_svg
from tqec.visualisation.computation.observable import get_observable_svg
from tqec.visualisation.computation.plaquette.base import SVGPlaquetteDrawer
from tqec.visualisation.configuration import DrawerConfiguration


def plaquette_grid_to_svg(
    grid: Sequence[Sequence[int]],
    drawers: FrozenDefaultDict[int, SVGPlaquetteDrawer],
    top_left_qubit: GridQubit,
    width: float,
    height: float,
    show_interaction_order: bool = True,
    show_hook_errors: bool = True,
    show_data_qubit_reset_measurements: bool = True,
    borders: tuple[float, float] = (0, 0),
    errors: Sequence[stim.ExplainedError] = tuple(),
    configuration: DrawerConfiguration = DrawerConfiguration(),
    observable: Observable | None = None,
) -> svg.Element:
    """Draws the provided plaquette grid and returns it as an SVG.

    The returned SVG is composed of 2 main parts:

    1. a first part that defines pieces of SVG that will be repeated across the drawing.
    2. a second part that uses the defined pieces by translating them to their expected place.

    Args:
        grid: a 2-dimensional array of integer corresponding to plaquettes indices that can be
            used to index the provided ``drawers``.
        drawers: a default dictionary containing a SVG drawer for each of the plaquette indices
            provided in ``grid``. If an index is not present, the default value is used.
        top_left_qubit: coordinates of the qubit at the very top-left of the
            visualisation canvas. Used to correctly offset qubit values from the
            provided ``errors``.
        width: width of the resulting SVG.
        height: height of the resulting SVG.
        show_interaction_order: if ``True``, numbers representing the timestep(s) at which each
            corner qubit is touched by a 2-qubit gates are added on the corners of each plqauette.
            Else, nothing is added.
        show_hook_errors: if ``True``, a dark line is added between the 2 qubits that may be
            vulnerable to a hook error. Else, nothing is added.
        show_data_qubit_reset_measurements: if ``True``, small squares/quarter circles will be added
            to each data qubit when they are reset/measured.
        borders: additional blank space reserved for border at the left/right and bottom/top. The
            width (resp. height) of the drawing area will be ``width - 2 * borders[0]`` (resp.
            ``height - 2 * borders[1]``).
        errors: a collection of errors that should be drawn on the resulting SVG.
        configuration: drawing configuration.
        observable: an observable that is being visualised. If provided, the SVG will be
            annotated with the observable information.

    Returns:
        a SVG element representing the provided plaquette grid.

    """
    if not grid:
        return svg.G()
    n, m = len(grid[0]), len(grid)
    xborder, yborder = borders
    pw, ph = (width - 2 * xborder) / n, (height - 2 * yborder) / m
    # Start by constructing an index of drawings that will be re-used.
    element_id_template = "plaquette_drawing_{}"
    drawing_lines: list[svg.Element] = [
        svg.Defs(
            elements=[
                drawer.draw(
                    element_id_template.format(i),
                    show_interaction_order,
                    show_hook_errors,
                    show_data_qubit_reset_measurements,
                    configuration,
                )
                for i, drawer in drawers.items()
            ]
            + (
                []
                if drawers.default_value is None
                else [
                    drawers.default_value.draw(
                        element_id_template.format("default"),
                        show_interaction_order,
                        show_hook_errors,
                        show_data_qubit_reset_measurements,
                        configuration,
                    )
                ]
            )
        )
    ]
    for i, line in enumerate(grid):
        y = ph * (i + yborder)
        for j, value in enumerate(line):
            x = pw * (j + xborder)
            if value == 0:
                continue
            key = str(value) if value in drawers else "default"
            template_id = element_id_template.format(key)
            drawing_lines.append(svg.Use(x=x, y=y, href=f"#{template_id}"))
    if errors:
        drawing_lines.append(get_errors_svg(errors, top_left_qubit, pw, ph, size=(pw + ph) / 12))
    if observable is not None:
        drawing_lines.append(get_observable_svg(observable, top_left_qubit, pw, ph))
    return svg.G(elements=drawing_lines)


def plaquette_grid_svg_viewer(
    grid: Sequence[Sequence[int]],
    drawers: FrozenDefaultDict[int, SVGPlaquetteDrawer],
    top_left_used_qubit: GridQubit,
    view_box_top_left_qubit: GridQubit | None = None,
    view_box_bottom_right_qubit: GridQubit | None = None,
    width: float | None = None,
    height: float | None = None,
    show_interaction_order: bool = True,
    show_hook_errors: bool = True,
    show_data_qubit_reset_measurements: bool = True,
    borders: tuple[float, float] = (0, 0),
    errors: Sequence[stim.ExplainedError] = tuple(),
    configuration: DrawerConfiguration = DrawerConfiguration(),
    observable: Observable | None = None,
) -> svg.SVG:
    """Draws the provided plaquette grid and returns it as an SVG.

    The returned SVG is composed of 2 main parts:

    1. a first part that defines pieces of SVG that will be repeated across the drawing.
    2. a second part that uses the defined pieces by translating them to their expected place.

    Args:
        grid: a 2-dimensional array of integer corresponding to plaquettes indices that can be
            used to index the provided ``drawers``.
        drawers: a default dictionary containing a SVG drawer for each of the plaquette indices
            provided in ``grid``. If an index is not present, the default value is used.
        top_left_used_qubit: coordinates of the top-left most qubit being used by the provided
            ``grid``. This is used to correctly offset qubit values from the provided ``errors``.
        view_box_top_left_qubit: top-left qubit of the drawing frame. Can be used to change the view
            box of the returned drawing. Contrary to ``width``, this parameter does **not** change
            the size of the plaquettes that are drawn, but changes the view box.
        view_box_bottom_right_qubit: bottom-right qubit of the drawing frame. Can be used to change
            the view box of the returned drawing. Contrary to ``height``, this parameter does
            **not** change the size of the plaquettes that are drawn, but changes the view box.
        width: width of the resulting SVG. If ``None``, defaults to ``len(grid[0])``.
        height: height of the resulting SVG. If ``None``, defaults to ``len(grid)``.
        show_interaction_order: if ``True``, numbers representing the timestep(s) at which each
            corner qubit is touched by a 2-qubit gates are added on the corners of each plqauette.
            Else, nothing is added.
        show_hook_errors: if ``True``, a dark line is added between the 2 qubits that may be
            vulnerable to a hook error. Else, nothing is added.
        show_data_qubit_reset_measurements: if ``True``, small squares/quarter circles will be added
            to each data qubit when they are reset/measured.
        borders: additional blank space reserved for border at the left/right and bottom/top. The
            width (resp. height) of the drawing area will be ``width - borders[0]`` (resp.
            ``height - borders[1]``).
        errors: a collection of errors that should be drawn on the resulting SVG.
        configuration: drawing configuration.
        observable: an observable that is being visualised. If provided, the SVG will be
            annotated with the observable information.

    Returns:
        a ``<svg>`` element that can be directly written to a ``.svg`` file and representing the
        provided plaquette grid.

    """
    if width is None and height is None:
        height, width = len(grid), len(grid[0])
    elif width is None:
        assert height is not None  # For type checkers
        width = len(grid[0]) * (height / len(grid))
    elif height is None:
        assert width is not None  # For type checkers
        height = len(grid) * (width / len(grid[0]))

    assert width is not None and height is not None

    # By default, if either the top-left or top-right qubits is not provided, we stick to the
    # [0, 0] x [width, height] viewbox.
    viewbox = svg.ViewBoxSpec(0, 0, width, height)
    if view_box_top_left_qubit is not None and view_box_bottom_right_qubit is not None:
        # Compute viewbox bounds top-left [x1, y1] and bottom-right [x2, y2] knowing that by
        # convention we have ``top_left_used_qubit`` at the coordinate ``(0, 0)``.
        width_between_qubits = width / (2 * len(grid[0]))
        height_between_qubits = height / (2 * len(grid))
        x1 = (view_box_top_left_qubit.x - top_left_used_qubit.x) * width_between_qubits
        x2 = (view_box_bottom_right_qubit.x - view_box_top_left_qubit.x) * width_between_qubits
        y1 = (view_box_top_left_qubit.y - top_left_used_qubit.y) * height_between_qubits
        y2 = (view_box_bottom_right_qubit.y - view_box_top_left_qubit.y) * height_between_qubits
        viewbox = svg.ViewBoxSpec(x1, y1, x2, y2)

    return svg.SVG(
        viewBox=viewbox,
        elements=[
            plaquette_grid_to_svg(
                grid,
                drawers,
                top_left_used_qubit,
                width,
                height,
                show_interaction_order,
                show_hook_errors,
                show_data_qubit_reset_measurements,
                borders,
                errors,
                configuration,
                observable,
            )
        ],
    )



"""Block graph that rotates boundary types by moving logical qubit in spacetime."""

from tqec.computation.block_graph import BlockGraph
from tqec.computation.cube import ZXCube
from tqec.utils.enums import Basis
from tqec.utils.position import Position3D


def move_rotation(observable_basis: Basis | None = None) -> BlockGraph:
    """Create a block graph for moving and rotating the spatial boundaries of a logical qubit.

    Args:
        observable_basis: The observable basis that the block graph can support. If None,
            the ports are left open. Otherwise, the ports are filled with the given basis.

    Returns:
        A :py:class:`~tqec.computation.block_graph.BlockGraph` instance representing
        the move-rotation operation.

    """
    g = BlockGraph("Move Rotation")
    nodes = [
        (Position3D(0, 0, 0), "P", "In"),
        (Position3D(0, 0, 1), "ZXX", ""),
        (Position3D(0, 1, 1), "ZZX", ""),
        (Position3D(1, 1, 1), "XZX", ""),
        (Position3D(1, 1, 2), "P", "Out"),
    ]
    for pos, kind, label in nodes:
        g.add_cube(pos, kind, label)

    pipes = [(0, 1), (1, 2), (2, 3), (3, 4)]
    for p0, p1 in pipes:
        g.add_pipe(nodes[p0][0], nodes[p1][0])

    if observable_basis == Basis.Z:
        g.fill_ports({"In": ZXCube.from_str("ZXZ"), "Out": ZXCube.from_str("XZZ")})
    elif observable_basis == Basis.X:
        g.fill_ports({"In": ZXCube.from_str("ZXX"), "Out": ZXCube.from_str("XZX")})

    return g



"""Block graph that represents a logical memory experiment."""

from tqec.computation.block_graph import BlockGraph
from tqec.utils.enums import Basis
from tqec.utils.position import Position3D


def memory(observable_basis: Basis = Basis.Z) -> BlockGraph:
    """Create a block graph with a single cube that represents a logical memory experiment.

    Args:
        observable_basis: The logical observable basis for the memory experiment.

    Returns:
        A :py:class:`~tqec.computation.block_graph.BlockGraph` instance.

    """
    g = BlockGraph(f"Logical {observable_basis} Memory Experiment")
    node_kind = f"ZX{observable_basis.value}"
    g.add_cube(Position3D(0, 0, 0), node_kind)
    return g



from ._version import __version__ as __version__
from .compile import compile_block_graph as compile_block_graph
from .computation.block_graph import BlockGraph as BlockGraph
from .utils import Basis as Basis
from .utils import NoiseModel as NoiseModel



import multiprocessing
from collections.abc import Callable, Iterable, Sequence
from math import isclose
from typing import TypeGuard

import sinter

from tqec.compile.compile import compile_block_graph
from tqec.compile.convention import FIXED_BULK_CONVENTION, Convention
from tqec.compile.detectors.database import DetectorDatabase
from tqec.computation.block_graph import BlockGraph
from tqec.computation.correlation import CorrelationSurface
from tqec.utils.exceptions import TQECError
from tqec.utils.noise_model import NoiseModel


def _is_only_floats(seq: Sequence[float | None]) -> TypeGuard[Sequence[float]]:
    return not any(elem is None for elem in seq)


def _is_sorted(seq: Sequence[int | float]) -> bool:
    if not seq:
        return True
    val = seq[0]
    for i in range(1, len(seq)):
        if seq[i] < val:
            return False
        val = seq[i]
    return True


def get_logical_error_rate_per_shot(
    stat: sinter.TaskStats, max_likelihood_factor: float = 1e3
) -> sinter.Fit:
    """Estimates the logical error rate per shot for the given ``stat``.

    Note:
        This function is heavily inspired from an internal function defined in
        ``sinter._plotting.plot_error_rate``.

    Args:
        stat: the statistics to use in order to compute the logical error rate
            per shot.
        max_likelihood_factor: Controls how wide the uncertainty highlight region
            around curves is. Must be 1 or larger. Hypothesis probabilities at
            most that many times as unlikely as the max likelihood hypothesis
            will be highlighted. Forwarded to ``sinter.fit_binomial``.

    Returns:
        The estimation of the logical error rate per shot with appropriate error
        bars.

    """
    result = sinter.fit_binomial(
        num_shots=stat.shots - stat.discards,
        num_hits=stat.errors,
        max_likelihood_factor=max_likelihood_factor,
    )
    if stat.errors == 0:
        result = sinter.Fit(low=result.low, high=result.high, best=float("nan"))

    return result


def binary_search_threshold(
    block_graph: BlockGraph,
    observable: CorrelationSurface,
    noise_model_factory: Callable[[float], NoiseModel],
    minp: float = 1e-5,
    maxp: float = 0.1,
    ks: Sequence[int] = (1, 2),
    atol: float = 1e-4,
    rtol: float = 1e-4,
    manhattan_radius: int = 2,
    convention: Convention = FIXED_BULK_CONVENTION,
    detector_database: DetectorDatabase | None = None,
    num_workers: int = multiprocessing.cpu_count(),
    max_shots: int = 10_000_000,
    max_errors: int = 5_000,
    decoders: Iterable[str] = ("pymatching",),
) -> tuple[float, dict[int, list[tuple[float, sinter.Fit]]]]:
    """Search the threshold value for the provided ``observable`` on the provided ``block_graph``.

    This function performs a binary search on the noise strength (called ``p`` in
    the following explanation) that is commonly called "threshold" in quantum
    error correction.

    The threshold is defined as the noise strength ``p`` below which increasing
    the code distance improves the logical error rate.

    Note:
        For high values of ``p``, the logical error-rate of an observable will
        be very close to ``0.5``, which means that the code is completely unable
        to correct the errors introduced by the operations it requires to work.
        In order to avoid this regime, the binary search will consider that any
        value of ``p`` that leads to logical error-rates above ``0.4`` for all
        the provided ``ks`` is too high (and so recurse the binary search in the
        first half of the interval).

    Warning:
        Small values for ``atol`` and ``rtol`` make very little sense here.

        Due to finite sampling, logical error-rate estimations will not be exact
        and will have error bars. This function does not take these error bars
        into account yet and only uses the best estimate.

        Setting ``atol`` or ``rtol`` to very small values will give the
        impression that the output estimate is very precise, but the noise
        floor imposed by ``max_shots`` and the associated sampling error will
        still be present, and might be higher than ``atol`` and ``rtol``.

    Args:
        block_graph: a representation of the QEC computation to find the
            threshold of.
        observable: the observable that should be considered to find the
            threshold.
        noise_model_factory: a callable that is used to instantiate a noise
            model from the values of ``p`` that are explored by this function.
        minp: lower bound for the threshold value.
        maxp: upper bound for the threshold value.
        ks: values of the scaling parameter `k` to use in order to generate the
            circuits.
        atol: absolute tolerance used to compare the lower and upper bounds
            during the binary search. The search ends when these bounds are
            considered equal w.r.t. the provided ``atol`` and ``rtol``.
        rtol: relative tolerance used to compare the lower and upper bounds
            during the binary search. The search ends when these bounds are
            considered equal w.r.t. the provided ``atol`` and ``rtol``.
        manhattan_radius: radius used to automatically compute detectors. The
            best value to set this argument to is the minimum integer such that
            flows generated from any given reset/measurement, from any plaquette
            at any spatial/temporal place in the QEC computation, do not
            propagate outside of the qubits used by plaquettes spatially located
            at maximum `manhattan_radius` plaquettes from the plaquette the
            reset/measurement belongs to (w.r.t. the Manhattan distance).
            Default to 2, which is sufficient for regular surface code. If
            negative, detectors are not computed automatically and are not added
            to the generated circuits.
        convention: convention used to generate the quantum circuits.
        detector_database: an instance to retrieve from / store in detectors
            that are computed as part of the circuit generation.
        num_workers: The number of worker processes to use.
        max_shots: Defaults to None (unused). Stops the sampling process
            after this many samples have been taken from the circuit.
        max_errors: Defaults to None (unused). Stops the sampling process
            after this many errors have been seen in samples taken from the
            circuit. The actual number sampled errors may be larger due to
            batching.
        decoders: Defaults to None (specified by each Task). The names of the
            decoders to use on each Task. It must either be the case that each
            Task specifies a decoder and this is set to None, or this is an
            iterable and each Task has its decoder set to None.

    Returns:
        A tuple containing an estimation of the threshold and a collection of all
        the logical error-rates computed while searching the threshold

    """
    compiled_graph = compile_block_graph(block_graph, convention, [observable])
    ks = tuple(sorted(ks))
    noiseless_circuits = [
        compiled_graph.generate_stim_circuit(
            k, manhattan_radius=manhattan_radius, detector_database=detector_database
        )
        for k in ks
    ]
    computed_logical_errors: dict[int, list[tuple[float, sinter.Fit]]] = {k: [] for k in ks}
    while not isclose(minp, maxp, rel_tol=rtol, abs_tol=atol):
        midp = (minp + maxp) / 2
        noise_model = noise_model_factory(midp)
        stats = sinter.collect(
            num_workers=num_workers,
            tasks=[
                sinter.Task(
                    circuit=noise_model.noisy_circuit(circ),
                    json_metadata={"d": 2 * k + 1, "r": 2 * k + 1, "p": midp},
                )
                for k, circ in zip(ks, noiseless_circuits)
            ],
            max_shots=max_shots,
            max_errors=max_errors,
            decoders=decoders,
            hint_num_tasks=len(ks),
        )
        logical_errors_fits: list[sinter.Fit] = [
            get_logical_error_rate_per_shot(stat)
            for stat in sorted(stats, key=lambda s: s.json_metadata["d"])
        ]
        # Update computed_logical_errors
        for i, k in enumerate(ks):
            computed_logical_errors[k].append((midp, logical_errors_fits[i]))
        logical_errors: list[float | None] = [lerr_fit.best for lerr_fit in logical_errors_fits]
        if not _is_only_floats(logical_errors):
            raise TQECError(
                "One of the computed logical errors is None. That likely means "
                "that you need more shots in order to have at least a few errors "
                "on all the simulations."
            )
        if all(lerr is not None and lerr > 0.4 for lerr in logical_errors):
            # The physical error rate is too high and the logical error rate
            # saturates around 0.5, so lower maxp
            maxp = midp
        elif _is_sorted([lerr for lerr in logical_errors if lerr is not None]):
            # The smaller k is, the smaller the logical error rate is. That means
            # that midp is above the threshold.
            maxp = midp
        elif _is_sorted([lerr for lerr in logical_errors if lerr is not None][::-1]):
            # The smaller k is, the larger the logical error rate is. That means
            # that midp is below the threshold.
            minp = midp
        else:
            # There is no clear ordering, that means that:
            # 1. the user provided 3 or more values for k,
            # 2. the current midp value is in the transition zone: some
            #    distances (i.e., values of k) are sub-threshold, others are
            #    still above.
            # Returning now seems like the best move.
            break
    return (minp + maxp) / 2, computed_logical_errors



from typing import Final

from tqec.plaquette.compilation.base import PlaquetteCompiler
from tqec.plaquette.compilation.passes.controlled_gate_basis import (
    ChangeControlledGateBasisPass,
)
from tqec.plaquette.compilation.passes.measurement_basis import (
    ChangeMeasurementBasisPass,
)
from tqec.plaquette.compilation.passes.reset_basis import ChangeResetBasisPass
from tqec.plaquette.compilation.passes.scheduling import ChangeSchedulePass
from tqec.plaquette.compilation.passes.sort_targets import SortTargetsPass
from tqec.plaquette.compilation.passes.transformer import ScheduleConstant
from tqec.utils.enums import Basis


def _add_hadamard(mergeable_instructions: frozenset[str]) -> frozenset[str]:
    return mergeable_instructions | frozenset(["H"])


CSSPlaquetteCompiler: Final[PlaquetteCompiler] = PlaquetteCompiler(
    "CSS",
    [
        # Move schedules to have an empty schedule for basis change after resets
        # and before measurements.
        ChangeSchedulePass({0: 0, 6: 8} | {i: i + 1 for i in range(1, 6)}),
        # Change reset basis when needed
        ChangeResetBasisPass(Basis.Z),
        # Change measurement basis when needed
        ChangeMeasurementBasisPass(Basis.Z),
        # Change controlled gate basis when needed (CZ -> CX)
        ChangeControlledGateBasisPass(Basis.X, ScheduleConstant(1), ScheduleConstant(7)),
        # Sort the instruction targets to normalize the circuits.
        SortTargetsPass(),
    ],
    mergeable_instructions_modifier=_add_hadamard,
)



"""View COLLADA model in html with the help of ``three.js``."""

import base64
import html
import pathlib
from string import Template


class _ColladaHTMLViewer:
    """Helper class to view COLLADA model in html with the help of ``three.js``.

    This can display a COLLADA model in IPython compatible environments with the
    implementation of ``_repr_html_`` method.

    """

    HTML_TEMPLATE = Template("""
<!doctype html>
<html>

<head>
  <meta charset="UTF-8" />
  <script type="importmap">
      {
        "imports": {
          "three": "https://unpkg.com/three@0.138.0/build/three.module.js",
          "three-orbitcontrols": "https://unpkg.com/three@0.138.0/examples/jsm/controls/OrbitControls.js",
          "three-collada-loader": "https://unpkg.com/three@0.138.0/examples/jsm/loaders/ColladaLoader.js"
        }
      }
    </script>
    <style>
      body {
        background-color: white;
      }
    </style>
</head>

<body>
  <a download="model.dae" id="model-download-link" \
     href="data:text/plain;base64,$MODEL_BASE64_PLACEHOLDER">
    Download 3D Model as .dae File
  </a>
  <br />Mouse Wheel = Zoom. Left Drag = Orbit. Right Drag = Strafe.
  <div id="scene-container" style="width: calc(100vw - 32px); height: calc(100vh - 64px)">
    JavaScript Blocked?
  </div>

  <script type="module">
    import * as THREE from "three";
    import {OrbitControls} from "three-orbitcontrols";
    import {ColladaLoader} from "three-collada-loader";

    const container = document.getElementById("scene-container");
    const downloadLink = document.getElementById("model-download-link");

    // Remove IDs to avoid cross-cell interactions (workaround for Jupyter notebook issue)
    container.removeAttribute("id");
    downloadLink.removeAttribute("id");

    container.textContent = "Loading viewer...";

    async function init() {
      try {
        container.textContent = "Loading model...";

        const modelDataUri = downloadLink.href;
        const collada = await new ColladaLoader().loadAsync(modelDataUri);

        container.textContent = "Loading scene...";

        // Create the scene
        const scene = new THREE.Scene();
        scene.background = new THREE.Color("#CBDFC6");

        // Set up lights
        setupLights(scene);

        // Prepare and add the model to the scene
        processModel(collada.scene);
        scene.add(collada.scene);

        // Set up the camera and controls
        const camera = new THREE.PerspectiveCamera(
          35,
          container.clientWidth / container.clientHeight,
          0.1,
          100000,
        );
        const controls = new OrbitControls(camera, container);

        // Adjust camera to fit the scene
        fitCameraToObject(camera, controls, scene);

        // Add axes helper
        const axesHelper = new THREE.AxesHelper(2);
        axesHelper.rotation.x = -Math.PI / 2;
        scene.add(axesHelper);

        // Set up renderer
        const renderer = new THREE.WebGLRenderer({antialias: true});
        setupRenderer(renderer, container, camera, scene);

        // Start rendering
        renderScene(renderer, scene, camera, controls, container);
      } catch (ex) {
        container.textContent = "Failed to show model. " + ex;
        console.error(ex);
      }
    }

    function setupLights(scene) {
      const ambientLight = new THREE.AmbientLight(0xffffff, 1.4);
      scene.add(ambientLight);

      const directionalLight = new THREE.DirectionalLight(0xffffff, 1.0);
      directionalLight.position.set(10, 10, 10);
      directionalLight.castShadow = true;
      scene.add(directionalLight);
    }

    function processModel(modelScene) {
      modelScene.traverse((node) => {
        if (node.isMesh) {
          node.material.side = THREE.DoubleSide;

          if (node.name.endsWith("correlation_surface")) {
            return;
          }
          // Add edges for better visibility
          const edgesGeometry = new THREE.EdgesGeometry(node.geometry);
          const edgesMaterial = new THREE.LineBasicMaterial({
            color: 0x000000,
          });
          const edgeMesh = new THREE.LineSegments(
            edgesGeometry,
            edgesMaterial,
          );
          node.add(edgeMesh);
        }
      });
    }

    function fitCameraToObject(camera, controls, scene) {
      let bounds = new THREE.Box3().setFromObject(scene);
      let mid = new THREE.Vector3(
        (bounds.min.x + bounds.max.x) * 0.5,
        (bounds.min.y + bounds.max.y) * 0.5,
        (bounds.min.z + bounds.max.z) * 0.5,
      );
      let boxPoints = [];
      for (let dx of [0, 0.5, 1]) {
        for (let dy of [0, 0.5, 1]) {
          for (let dz of [0, 0.5, 1]) {
            boxPoints.push(
              new THREE.Vector3(
                bounds.min.x + (bounds.max.x - bounds.min.x) * dx,
                bounds.min.y + (bounds.max.y - bounds.min.y) * dy,
                bounds.min.z + (bounds.max.z - bounds.min.z) * dz,
              ),
            );
          }
        }
      }
      let isInView = (p) => {
        p = new THREE.Vector3(p.x, p.y, p.z);
        p.project(camera);
        return Math.abs(p.x) < 1 && Math.abs(p.y) < 1 && p.z >= 0 && p.z < 1;
      };
      let unit = new THREE.Vector3(0.3, 0.4, 1.8);
      unit.normalize();
      let setCameraDistance = (d) => {
        controls.target.copy(mid);
        camera.position.copy(mid);
        camera.position.addScaledVector(unit, d);
        controls.update();
        return boxPoints.every(isInView);
      };

      let maxDistance = 1;
      for (let k = 0; k < 20; k++) {
        if (setCameraDistance(maxDistance)) {
          break;
        }
        maxDistance *= 2;
      }
      let minDistance = maxDistance;
      for (let k = 0; k < 20; k++) {
        minDistance /= 2;
        if (!setCameraDistance(minDistance)) {
          break;
        }
      }
      for (let k = 0; k < 20; k++) {
        let mid = (minDistance + maxDistance) / 2;
        if (setCameraDistance(mid)) {
          maxDistance = mid;
        } else {
          minDistance = mid;
        }
      }
      setCameraDistance(maxDistance);
    }

    function setupRenderer(renderer, container, camera, scene) {
      container.textContent = "";
      renderer.setSize(container.clientWidth, container.clientHeight);
      renderer.setPixelRatio(window.devicePixelRatio);
      renderer.physicallyCorrectLights = true;
      renderer.toneMapping = THREE.ACESFilmicToneMapping;
      renderer.toneMappingExposure = 1.2;
      renderer.outputEncoding = THREE.sRGBEncoding;
      renderer.setClearColor(0xdedede);
      container.appendChild(renderer.domElement);
    }

    function renderScene(renderer, scene, camera, controls, container) {
      const render = () => {
        renderer.render(scene, camera);
      };

      render();

      new ResizeObserver(() => {
        const width = container.clientWidth;
        const height = container.clientHeight;
        camera.aspect = width / height;
        camera.updateProjectionMatrix();
        renderer.setSize(width, height);
        render();
      }).observe(container);

      controls.addEventListener("change", render);
    }

    init();
  </script>
</body>

</html>
""")

    def __init__(self, filepath_or_bytes: str | pathlib.Path | bytes) -> None:  # pragma: no cover
        if isinstance(filepath_or_bytes, bytes):
            collada_bytes = filepath_or_bytes
        else:
            with open(filepath_or_bytes, "rb") as file:
                collada_bytes = file.read()
        collada_base64 = base64.b64encode(collada_bytes).decode("utf-8")
        self.html_str = self.HTML_TEMPLATE.substitute(MODEL_BASE64_PLACEHOLDER=collada_base64)

    def _repr_html_(self) -> str:  # pragma: no cover
        framed = """<iframe style="width: 100%; height: 300px; overflow: hidden; resize: both; """
        framed += """border: 1px dashed gray;" frameBorder="0" """
        framed += f"""srcdoc="{html.escape(self.html_str, quote=True)}"></iframe>"""
        return framed

    def __str__(self) -> str:
        return self.html_str  # pragma: no cover


def display_collada_model(  # pragma: no cover
    filepath_or_bytes: str | pathlib.Path | bytes,
    write_html_filepath: str | pathlib.Path | None = None,
) -> _ColladaHTMLViewer:
    """Display the 3D COLLADA model from a Collada DAE file in IPython compatible environments.

    If ``write_html_filepath`` is provided (i.e., not ``None``), this function writes the generated
    HTML content to a file.

    The implementation of this function references the code snippet from the
    ``stim.Circuit().diagram()`` method.

    Args:
        filepath_or_bytes: The input dae file path or bytes of the dae file.
        write_html_filepath: The output html file path to write the generated html content.
          Default is None.

    Returns:
        A helper class to display the 3D model, which implements the ``_repr_html_`` method and
        can be directly displayed in IPython compatible environments.

    """
    helper = _ColladaHTMLViewer(filepath_or_bytes)

    if write_html_filepath is not None:
        with open(write_html_filepath, "w") as file:
            file.write(str(helper))

    return helper



from __future__ import annotations

from collections.abc import Iterator
from enum import Flag, auto

from tqec.computation.block_graph import BlockGraph
from tqec.computation.cube import Cube
from tqec.utils.exceptions import TQECError


class SpatialArms(Flag):
    NONE = 0
    UP = auto()
    RIGHT = auto()
    DOWN = auto()
    LEFT = auto()

    @staticmethod
    def from_cube_in_graph(cube: Cube, graph: BlockGraph) -> SpatialArms:
        """Return the spatial arms of a cube in a block graph."""
        if not cube.is_spatial:
            return SpatialArms.NONE  # pragma: no cover
        pos = cube.position
        if pos not in graph or graph[pos] != cube:
            raise TQECError(f"Cube {cube} is not in the graph.")
        spatial_arms = SpatialArms.NONE
        for flag, shift in SpatialArms.get_map_from_arm_to_shift().items():
            if graph.has_pipe_between(pos, pos.shift_by(*shift)):
                spatial_arms |= flag
        return spatial_arms

    @property
    def has_spatial_arm_in_both_dimensions(self) -> bool:
        """Checks if the spatial arms have arms in both spatial dimensions."""
        return (SpatialArms.DOWN in self or SpatialArms.UP in self) and (
            SpatialArms.LEFT in self or SpatialArms.RIGHT in self
        )

    @classmethod
    def get_map_from_arm_to_shift(cls) -> dict[SpatialArms, tuple[int, int]]:
        """Return a mapping from any single arm to 2-dimensional shifts.

        The returned shifts are the shift needed to get the cube at the other end of the associated
        single arm.

        If a block ``B`` has a LEFT arm, the other end of the arm is the coordinates of ``B``
        shifted by ``(-1, 0)``. Hence,
        ``SpatialArms.get_map_from_arm_to_shift()[SpatialArms.LEFT] == (-1 0)``.

        Warning:
            In TQEC convention, the ``Y`` axis is pointing **downwards**. That means that UP is
            linked to ``(0, -1)`` in the returned dictionary.

        Returns:
            a mapping from any single arm to the shift needed to get the cube at the other end of
            that arm.

        """
        return {
            cls.UP: (0, -1),
            cls.RIGHT: (1, 0),
            cls.DOWN: (0, 1),
            cls.LEFT: (-1, 0),
        }

    @staticmethod
    def I_shaped_arms() -> list[SpatialArms]:
        """Return the 2 arm combinations that form a I-shape.

        The 2 combinations are UP | DOWN and LEFT | RIGHT.

        """
        return [
            SpatialArms.DOWN | SpatialArms.UP,
            SpatialArms.LEFT | SpatialArms.RIGHT,
        ]

    @staticmethod
    def L_shaped_arms() -> list[SpatialArms]:
        """Return the 4 arm combinations that form a L-shape.

        The 4 combinations are DOWN | LEFT, DOWN | RIGHT, UP | LEFT and UP | RIGHT.

        """
        return [  # pragma: no cover
            SpatialArms.DOWN | SpatialArms.LEFT,
            SpatialArms.DOWN | SpatialArms.RIGHT,
            SpatialArms.UP | SpatialArms.LEFT,
            SpatialArms.UP | SpatialArms.RIGHT,
        ]

    @staticmethod
    def T_shaped_arms() -> list[SpatialArms]:
        """Return the 4 arm combinations that form a T-shape."""
        return [  # pragma: no cover
            SpatialArms.DOWN | SpatialArms.LEFT | SpatialArms.UP,
            SpatialArms.LEFT | SpatialArms.UP | SpatialArms.RIGHT,
            SpatialArms.UP | SpatialArms.RIGHT | SpatialArms.DOWN,
            SpatialArms.RIGHT | SpatialArms.DOWN | SpatialArms.LEFT,
        ]

    @staticmethod
    def X_shaped_arms() -> list[SpatialArms]:
        """Return the only arm combinations that form a X-shape (i.e., all the arms)."""
        return [  # pragma: no cover
            SpatialArms.DOWN | SpatialArms.LEFT | SpatialArms.UP | SpatialArms.RIGHT
        ]

    @staticmethod
    def single_arms() -> list[SpatialArms]:
        """Return the 4 possible single arms."""
        return [
            SpatialArms.UP,
            SpatialArms.RIGHT,
            SpatialArms.DOWN,
            SpatialArms.LEFT,
        ]

    def __len__(self) -> int:
        return sum(arm in self for arm in SpatialArms.single_arms())

    def __iter__(self) -> Iterator[SpatialArms]:
        yield from (arm for arm in SpatialArms.single_arms() if arm in self)  # pragma: no cover

    def __repr__(self) -> str:
        if self == SpatialArms.NONE:
            return f"{SpatialArms.__name__}.NONE"
        return " | ".join(f"{SpatialArms.__name__}.{arm.name}" for arm in self)



"""Collection of strategies for synthesizing a block graph from a ZX graph."""

from collections.abc import Mapping
from enum import Enum
from typing import Any

from pyzx.graph.graph_s import GraphS

from tqec.computation.block_graph import BlockGraph
from tqec.interop.pyzx.positioned import PositionedZX
from tqec.interop.pyzx.synthesis.positioned import positioned_block_synthesis
from tqec.utils.position import Position3D


class SynthesisStrategy(Enum):
    """Strategy for synthesizing a block graph from a ZX graph."""

    POSITIONED = "POSITIONED"
    """Mapping from a :py:class:`~tqec.interop.pyzx.positioned_graph.PositionedZX` instance to a
    block graph.

    This strategy requires specifying the 3D positions of each vertex explicitly in the ZX graph.
    Then the conversion converts each vertex by looking at its nearest neighbors to infer the cube
    kind. The conversion maps each vertex to a cube in the block graph and each edge to a pipe
    connecting the corresponding cubes.

    """


def block_synthesis(
    zx_graph: GraphS,
    strategy: SynthesisStrategy = SynthesisStrategy.POSITIONED,
    *,
    positions: Mapping[int, Position3D] | None = None,
    **_kwargs: dict[str, Any],
) -> BlockGraph:
    """Perform block synthesis translating an instance of ``pyzx.GraphS`` to a :class:`.BlockGraph`.

    Args:
        zx_graph: a ZX graph to transform into a :class:`.BlockGraph`.
        strategy: strategy used to perform the translation.
        positions: explicit block positioning for each node in the provided ``zx_graph``. Only used
            when ``strategy == SynthesisStrategy.POSITIONED``.

    Raises:
        ValueError: if ``strategy == SynthesisStrategy.POSITIONED`` and ``positions`` is not
            provided.

    Returns:
        an instance of :class:`.BlockGraph` representing the provided ``zx_graph``.

    """
    match strategy:
        case SynthesisStrategy.POSITIONED:
            if positions is None:
                raise ValueError(
                    "The POSITIONED strategy requires specifying positions, but None was given."
                )
            g = PositionedZX(zx_graph, positions)
            return positioned_block_synthesis(g)



r"""Defines several scalable arrays of numbers used to tile plaquettes.

This module defines the :py:class:`~tqec.templates.base.Template` interface that
should be implemented by subclasses implementing different templates.

Terminology used
================

The **template** terminology is used to describe an object (instance of a subclass
of the :py:class:`~tqec.templates.base.Template` class) that can be
"instantiated" into an array of integers from an input integer scaling parameter
:math:`k`.

A **subtemplate** is a constant-size sub-array of a template instantiation. For the
moment, a subtemplate is always a square sub-array with odd-sized sides.

The **scaling parameter** :math:`k` directly relates to the code distance :math:`d`
(at least for regular surface code) where :math:`d = 2 k + 1`.

Templates
=========

As noted in the terminology section, a "template" is defined as something that
can be "instantiated", i.e., that can generate an array of integer from a given
scaling parameter :math:`k`.

Templates should all inherit from the base :py:class:`~tqec.templates.base.Template`
class that defines a few abstract methods that have to be overridden by child
classes. The most canonical example of a template is implemented by the
:py:class:`~tqec.templates.qubit.QubitTemplate` class: it represents the
arrangement of plaquettes needed to make a logical qubit.

The following code demonstrates quite well what a template is supposed to be.

.. code-block:: python

    from tqec.templates.display import display_template
    from tqec.templates.qubit import QubitTemplate

    template = QubitTemplate()
    display_template(template, 2)

It outputs

.. code-block:: text

    1   5   6   5   6   2
    7   9   10  9   10  11
    8   10  9   10  9   12
    7   9   10  9   10  11
    8   10  9   10  9   12
    3   13  14  13  14  4

The template instantiation is an array of size :math:`6 \times 6`, filled with
integers. Each unique integer will eventually either be linked to a plaquette or
be left empty. Let's try to clean up a little bit the array by setting the
values ``1``, ``2``, ``3``, ``4``, ``5``, ``8``, ``12`` and ``14`` to the value
``0`` that means "no plaquette here" for all templates (and that is represented
by ``.`` for visibility).

.. code-block:: python

    k = 2
    removed_plaquettes = {1, 2, 3, 4, 5, 8, 12, 14}
    display_template(
        template, k, [i if i not in removed_plaquettes else 0 for i in range(1, 15)]
    )

that outputs

.. code-block:: text

    .   .   6   .   6   .
    7   9   10  9   10  11
    .   10  9   10  9   .
    7   9   10  9   10  11
    .   10  9   10  9   .
    .   13  .   13  .   .

The above array looks very much like a logical qubit arrangement as can be seen in
most of the papers using the surface code and below!

.. image:: /media/api/tqec/templates/logical_qubit.png

But that is only a distance
:math:`5 = 2 \times 2 + 1 = 2 k + 1` code. Let's try to scale that up:

.. code-block:: python

    k = 5
    display_template(
        template, k, [i if i not in removed_plaquettes else 0 for i in range(1, 15)]
    )

outputs what we would expect: the "template" of plaquettes for a logical qubit
using a distance :math:`d = 2 k + 1 = 2 \times 5 + 1 = 11` code:

.. code-block:: text

    .   .   6   .   6   .   6   .   6   .   6   .
    7   9   10  9   10  9   10  9   10  9   10  11
    .   10  9   10  9   10  9   10  9   10  9   .
    7   9   10  9   10  9   10  9   10  9   10  11
    .   10  9   10  9   10  9   10  9   10  9   .
    7   9   10  9   10  9   10  9   10  9   10  11
    .   10  9   10  9   10  9   10  9   10  9   .
    7   9   10  9   10  9   10  9   10  9   10  11
    .   10  9   10  9   10  9   10  9   10  9   .
    7   9   10  9   10  9   10  9   10  9   10  11
    .   10  9   10  9   10  9   10  9   10  9   .
    .   13  .   13  .   13  .   13  .   13  .   .

This ability to scale to arbitrarily large values of :math:`k` (as long as your
computer can store a matrix of shape :math:`(2k+1, 2k+1)`) is exactly the reason
of existence of the :py:class:`~tqec.templates.base.Template` class.

This module also includes a few sub-classes implementing the
:py:class:`~tqec.templates.base.Template` interface:

1. :py:class:`~tqec.templates.qubit.QubitTemplate` that has already been used in
   documentation and that represents a logical qubit.
2. :py:class:`~tqec.templates.qubit.QubitSpatialCubeTemplate` that represents a
   a logical qubit with all the spatial boundaries in the same basis.
3. :py:class:`~tqec.templates.layout.LayoutTemplate` that represent an arbitrary
   layout of other templates arranged on a regular grid.

Sub-templates
=============

This module also implements helpers to list and represent sub-templates. A
sub-template is basically a sub-array of a given template instantiation.
Sub-templates are all of known constant size and cannot be scaled. Sub-templates
are represented as an array of integers with a square shape (i.e., 2-dimensional
with the same number of elements in each dimension) with odd-sized sides.

The module :py:mod:`tqec.templates.subtemplates` provides functions to compute
all the subtemplates for given template instantiation and subtemplate size along
with some data-structure to represent efficiently this collection of
sub-templates.

Displaying
==========

Templates are basically scalable arrays of integers. The module
:py:mod:`tqec.templates.display` provides a few functions to display a pretty and
human readable representation of a given template instantiation. These functions
are used in the documentation about templates above, and so their output can be
observed above.

"""



from collections.abc import Sequence

from tqec.compile.observables.abstract_observable import (
    CubeWithArms,
    PipeWithArms,
    PipeWithObservableBasis,
)
from tqec.compile.observables.builder import Coordinates2D, ObservableBuilder
from tqec.compile.specs.enums import SpatialArms
from tqec.computation.cube import ZXCube
from tqec.utils.enums import Basis, Orientation
from tqec.utils.position import Direction3D, PlaquetteShape2D, SignedDirection3D


def build_regular_cube_top_readout_qubits(
    shape: PlaquetteShape2D, observable_orientation: Orientation
) -> Sequence[Coordinates2D]:
    """Build the qubit coordinates for the middle line of the top face of a regular cube."""
    if observable_orientation == Orientation.HORIZONTAL:
        return [(x, shape.y // 2) for x in range(1, shape.x)]
    return [(shape.x // 2, y) for y in range(1, shape.y)]


def build_spatial_cube_top_readout_qubits(
    shape: PlaquetteShape2D, arms: SpatialArms, observable_basis: Basis
) -> Sequence[Coordinates2D]:
    """Build the qubit coordinates for a straight or bent middle line of the top face of a spatial
    cube.
    """
    assert len(arms) == 2
    half_x, half_y = shape.x // 2, shape.y // 2

    if arms == SpatialArms.LEFT | SpatialArms.RIGHT:
        return [(x, half_y) for x in range(1, shape.x)]
    if arms == SpatialArms.UP | SpatialArms.DOWN:
        return [(half_x, y) for y in range(1, shape.y)]
    if arms == SpatialArms.LEFT | SpatialArms.UP:
        qubits = [(x, half_y) for x in range(1, half_x)] + [(half_x, y) for y in range(1, half_y)]
        if observable_basis == Basis.Z:
            qubits.append((half_x, half_y))
        return qubits
    if arms == SpatialArms.DOWN | SpatialArms.RIGHT:
        qubits = [(x, half_y) for x in range(shape.x - 1, half_x, -1)] + [
            (half_x, y) for y in range(shape.y - 1, half_y, -1)
        ]
        if observable_basis == Basis.Z:
            qubits.append((half_x, half_y))
        return qubits
    if arms == SpatialArms.UP | SpatialArms.RIGHT:
        qubits = [(x, half_y) for x in range(shape.x - 1, half_x, -1)] + [
            (half_x, y) for y in range(1, half_y)
        ]
        if observable_basis == Basis.X:
            qubits.append((half_x, half_y))
        return qubits
    # arms == SpatialArms.LEFT | SpatialArms.DOWN:
    qubits = [(x, half_y) for x in range(1, half_x)] + [
        (half_x, y) for y in range(shape.y - 1, half_y, -1)
    ]
    if observable_basis == Basis.X:
        qubits.append((half_x, half_y))
    return qubits


def build_cube_top_readout_qubits(
    shape: PlaquetteShape2D, cube: CubeWithArms
) -> Sequence[Coordinates2D]:
    """Build the qubit coordinates whose measurements will be included in the observable on the top
    face of a cube.
    """
    if not cube.cube.is_spatial:
        kind = cube.cube.kind
        assert isinstance(kind, ZXCube)
        # Determine the middle line orientation based on the cube kind.
        # Since the basis of the top face decides the measurement basis of the data
        # qubits, i.e. the logical operator basis. We only need to find the spatial
        # boundaries that the logical operator can be attached to.
        obs_orientation = Orientation.VERTICAL if kind.y == kind.z else Orientation.HORIZONTAL
        return build_regular_cube_top_readout_qubits(shape, obs_orientation)
    kind = cube.cube.kind
    assert isinstance(kind, ZXCube)
    observable_basis = kind.z
    return build_spatial_cube_top_readout_qubits(shape, cube.arms, observable_basis)


def build_pipe_top_readout_qubits(
    shape: PlaquetteShape2D, direction: Direction3D
) -> Sequence[Coordinates2D]:
    """Build the qubit coordinates whose measurements will be included in the observable on the top
    face of a pipe.
    """
    assert direction != Direction3D.Z
    if direction == Direction3D.X:
        return [(shape.x, shape.y // 2)]
    return [(shape.x // 2, shape.y)]


def build_regular_cube_bottom_stabilizer_qubits(
    shape: PlaquetteShape2D,
    connect_to: SignedDirection3D,
    stabilizer_basis: Basis,
) -> Sequence[Coordinates2D]:
    """Build the stabilizer measurement coordinates that will be included in the observable
    spanning half of the bottom face of a regular cube.
    """
    stabilizers: list[tuple[float, float]] = []
    xy_sum_parity = 0 if stabilizer_basis == Basis.Z else 1
    # We calculate the qubits for the connect_to=SignedDirection3D(Direction3D.X, True) case
    # and rotate to get the correct orientation.
    for i in range(shape.x // 2, shape.x):
        for j in range(shape.y):
            if (i + j) % 2 != xy_sum_parity:
                continue
            x = i + 0.5
            y = j + 0.5
            # reflect the coordinates along the middle line of the cube
            # if the direction is along Y to preserve the checkerboard parity
            if connect_to.direction == Direction3D.Y:
                y = shape.y - y
            stabilizers.append((x, y))

    # rotate all coordinates around the block center:
    # rx = cx + a * (x - cx) - b * (y - cy)
    # ry = cy + b * (x - cx) + a * (y - cy)
    # in which (cx, cy) is the center of the block, a = cos(theta), b = sin(theta)
    # and theta is the angle of rotation.
    a, b = 1, 0
    match connect_to:
        case SignedDirection3D(Direction3D.X, False):
            a, b = -1, 0
        case SignedDirection3D(Direction3D.Y, True):
            a, b = 0, 1
        case SignedDirection3D(Direction3D.Y, False):
            a, b = 0, -1
        case _:
            pass
    cx, cy = shape.x // 2, shape.y // 2
    return [
        (
            cx + a * (x - cx) - b * (y - cy),
            cy + b * (x - cx) + a * (y - cy),
        )
        for x, y in stabilizers
    ]


def build_spatial_cube_bottom_stabilizer_qubits(
    shape: PlaquetteShape2D, stabilizer_basis: Basis
) -> Sequence[Coordinates2D]:
    """Build the stabilizer measurement coordinates that will be included in the observable
    spanning the bottom face of a spatial cube.
    """
    xy_sum_parity = 0 if stabilizer_basis == Basis.Z else 1
    return [
        (i + 0.5, j + 0.5)
        for i in range(shape.x)
        for j in range(shape.y)
        if (i + j) % 2 == xy_sum_parity
    ]


def build_cube_bottom_stabilizer_qubits(
    shape: PlaquetteShape2D, cube: CubeWithArms
) -> Sequence[Coordinates2D]:
    """Build the stabilizer measurement coordinates that will be included in the observable on the
    bottom face of a cube.
    """
    assert cube.cube.is_spatial
    kind = cube.cube.kind
    assert isinstance(kind, ZXCube)
    return build_spatial_cube_bottom_stabilizer_qubits(shape, kind.x)


def build_pipe_bottom_stabilizer_qubits(
    shape: PlaquetteShape2D, pipe: PipeWithArms
) -> Sequence[Coordinates2D]:
    """Build the stabilizer measurement coordinates that will be included in the observable on the
    bottom face of a pipe.

    It includes the bottom stabilizers of the connected cubes.

    """
    direction = pipe.pipe.direction

    qubits: list[Coordinates2D] = []
    for cube in pipe.pipe:
        kind = cube.kind
        assert isinstance(kind, ZXCube)
        is_u = cube == pipe.pipe.u
        if cube.is_spatial:
            stabilizers = build_spatial_cube_bottom_stabilizer_qubits(shape, kind.x)
        else:
            connect_to = SignedDirection3D(direction, is_u)
            stabilizer_basis = kind.get_basis_along(Direction3D(1 - direction.value))
            stabilizers = build_regular_cube_bottom_stabilizer_qubits(
                shape, connect_to, stabilizer_basis
            )
        # the local coordinates is of cube u, therefore, we need to shift the coordinates
        # of the stabilizers within cube v
        if not is_u:
            stabilizers = [
                (
                    s[0] + shape.x if direction == Direction3D.X else s[0],
                    s[1] + shape.y if direction == Direction3D.Y else s[1],
                )
                for s in stabilizers
            ]
        qubits.extend(stabilizers)
    return qubits


def _build_pipe_temporal_hadamard_qubits_impl(
    shape: PlaquetteShape2D, observable_basis: Basis, z_orientation: Orientation
) -> Sequence[Coordinates2D]:
    # observable is horizontal
    if (observable_basis == Basis.X) ^ (z_orientation == Orientation.HORIZONTAL):
        if (shape.x % 4 == 0) ^ (observable_basis == Basis.Z):
            return [(shape.x - 0.5, shape.y // 2 + 0.5)]
        return []
    # observable is vertical
    if (shape.y % 4 == 0) ^ (observable_basis == Basis.Z):
        return [(shape.x // 2 + 0.5, shape.y - 0.5)]
    return []


def build_pipe_temporal_hadamard_qubits(
    shape: PlaquetteShape2D, pipe: PipeWithObservableBasis
) -> Sequence[Coordinates2D]:
    """Build the stabilizer measurement coordinates that will be included in the observable at the
    realignment layer of a temporal Hadamard pipe.
    """
    pipe_kind = pipe.pipe.kind
    observable_basis = pipe.observable_basis
    z_orientation = (
        Orientation.VERTICAL
        if pipe_kind.get_basis_along(Direction3D.Y) == Basis.Z
        else Orientation.HORIZONTAL
    )
    return _build_pipe_temporal_hadamard_qubits_impl(shape, observable_basis, z_orientation)


FIXED_BULK_OBSERVABLE_BUILDER = ObservableBuilder(
    cube_top_readouts_builder=build_cube_top_readout_qubits,
    pipe_top_readouts_builder=lambda shape, pipe: build_pipe_top_readout_qubits(
        shape, pipe.pipe.direction
    ),
    cube_bottom_stabilizers_builder=build_cube_bottom_stabilizer_qubits,
    pipe_bottom_stabilizers_builder=build_pipe_bottom_stabilizer_qubits,
    pipe_temporal_hadamard_builder=build_pipe_temporal_hadamard_qubits,
)



"""Defines :class:`CorrelationSurface` and functions to find and build them from a ZX graph."""

from __future__ import annotations

import multiprocessing
import operator
from collections import ChainMap
from collections.abc import (
    Callable,
    Generator,
    Iterable,
    Iterator,
    MutableMapping,
    Sequence,
)
from copy import copy
from functools import cache, partial, reduce
from itertools import (
    accumulate,
    chain,
    combinations,
    pairwise,
    product,
    repeat,
    starmap,
)
from typing import Any, TypeVar

from pyzx.graph.graph_s import GraphS
from pyzx.utils import VertexType
from typing_extensions import Self

from tqec.computation.correlation import CorrelationSurface, ZXEdge, ZXNode
from tqec.interop.pyzx.utils import (
    is_hadamard,
    vertex_type_to_pauli,
    zx_to_pauli,
)
from tqec.utils.enums import Basis, Pauli
from tqec.utils.exceptions import TQECError


class _CorrelationSurfaceBase(MutableMapping[int, dict[int, Pauli]]):
    """Correlation surface represented as Pauli operators on half-edges."""

    def _add_pauli_to_edge(
        self, edge: tuple[int, int], pauli: Pauli, edge_is_hadamard: bool
    ) -> Self:
        """Add Pauli operators to both ends of the given edge."""
        for (u, v), p in zip(
            (edge, edge[::-1]),
            (pauli, pauli.flipped(edge_is_hadamard)),
        ):
            edges = self.setdefault(u, {})
            edges[v] = p
        return self

    def _paulis_at_nodes(self, nodes: Iterable[int]) -> Iterable[Pauli]:
        """Get the Pauli operators at the given nodes."""
        return chain.from_iterable(self[n].values() for n in nodes)

    def _signature_at_nodes(
        self,
        nodes: Iterable[int],
        func: Callable[[Pauli], int] = lambda p: p.value,
        bit_length: int = 2,
    ) -> int:
        """Compute the signature at the given nodes using the provided function."""
        ints = self._paulis_at_nodes(nodes)
        return _concat_ints_as_bits(map(func, ints), bit_length)

    def _validate_node(
        self, node: int, node_basis: Pauli, has_unconnected_neighbors: bool
    ) -> int | tuple[Pauli, bool]:
        """Return the broadcast Pauli and passthrough parity if valid or the syndrome otherwise."""
        paulis = list(self._paulis_at_nodes([node]))
        passthrough_parity = node_basis in reduce(operator.xor, paulis)
        valid = True
        broadcast_basis = node_basis.flipped()
        syndrome = [broadcast_basis in pauli for pauli in paulis]
        if all(syndrome):
            broadcast_pauli = broadcast_basis
        elif not any(syndrome):
            broadcast_pauli = Pauli.I
        else:  # invalid broadcast
            valid = False
        if not has_unconnected_neighbors:
            syndrome.append(passthrough_parity)
            if passthrough_parity:  # invalid passthrough
                valid = False
        if valid:
            return broadcast_pauli, passthrough_parity  # type: ignore (broadcast_pauli is always bound here)
        return _concat_ints_as_bits(syndrome, 1)

    def _to_immutable_public_representation(self, zx_graph: GraphS) -> CorrelationSurface:
        """Convert to the public representation of correlation surface."""
        span: list[ZXEdge] = []
        zx_nodes: dict[tuple[int, Basis], ZXNode] = {}
        bases = list(Basis)
        for u, v in zx_graph.edges():
            pauli_u = self[u][v]
            pauli_v = self[v][u]
            edge_is_hadamard = is_hadamard(zx_graph, (u, v))
            for xz_u, xz_v in product(Pauli.iter_xz(), repeat=2):
                if (edge_is_hadamard ^ (xz_u is xz_v)) and xz_u in pauli_u and xz_v in pauli_v:
                    basis_u = bases[xz_u.value >> 1]
                    basis_v = bases[xz_v.value >> 1]
                    node_u = zx_nodes.setdefault((u, basis_u), ZXNode(u, basis_u))
                    node_v = zx_nodes.setdefault((v, basis_v), ZXNode(v, basis_v))
                    span.append(ZXEdge.sorted(node_u, node_v))
        return CorrelationSurface(frozenset(span))


_CorrelationSurfaceType = TypeVar("_CorrelationSurfaceType", bound="_CorrelationSurfaceBase")


# Due to subtle differences in how generics and overloads are defined in the stubs, the type
# checker will say that dict.get is not a strictly valid replacement for MutableMapping.get.
class _CorrelationSurface(dict[int, dict[int, Pauli]], _CorrelationSurfaceBase):  # type: ignore
    """Correlation surface represented as Pauli operators on half-edges."""

    ...


class _CorrelationSurfaceView(ChainMap[int, dict[int, Pauli]], _CorrelationSurfaceBase):
    """A view of multiple correlation surfaces representing a single one."""

    def __setitem__(self, key, value):
        for mapping in self.maps:
            if key in mapping:
                mapping[key] = value
                return
        self.maps[0][key] = value

    def __delitem__(self, key):
        for mapping in self.maps:
            if key in mapping:
                del mapping[key]
                return
        raise KeyError(key)


def _xor_correlation_surfaces(
    correlation_surfaces: list[_CorrelationSurfaceBase],
) -> _CorrelationSurface:
    """XOR multiple correlation surfaces together."""
    # This method is deliberately written in this verbose manner, rather than more concisely with
    # zip, reduce, etc., to avoid the slight overhead, which becomes noticeable at a large scale.
    result = _CorrelationSurface()
    others = correlation_surfaces[1:]
    for v, neighbors in correlation_surfaces[0].items():
        result_neighbors = result.setdefault(v, {})
        other_neighbor_rows = [cs[v] for cs in others]
        for n, pauli in neighbors.items():
            for neighbor_row in other_neighbor_rows:
                pauli ^= neighbor_row[n]  # noqa: PLW2901
            result_neighbors[n] = pauli
    return result


def _partition_graph_from_vertices(
    zx_graph: GraphS, vertices_list: Sequence[set[int]], add_cut_edge_as_boundary_node: bool = False
) -> tuple[list[GraphS], list[tuple[dict[int, tuple[int, int]], dict[int, tuple[int, int]]]]]:
    """Create subgraphs from given sets of vertices; optionally add boundary vertices at cuts."""
    # Needs to be imported here to avoid pulling pyzx when importing this module.
    from pyzx.graph.graph_s import GraphS  # noqa: PLC0415

    subgraphs = []
    cut_edges_map = {}
    added_vertices_list = []
    for vertices in vertices_list:
        subgraph = GraphS()
        input_vertices, output_vertices = {}, {}
        for v in vertices:
            subgraph.add_vertex_indexed(v)
            subgraph.set_type(v, zx_graph.type(v))
            subgraph.set_phase(v, zx_graph.phase(v))
        for v in vertices:
            for u in zx_graph.neighbors(v):
                if u in vertices:
                    if not subgraph.connected(u, v):
                        subgraph.add_edge((u, v), zx_graph.edge_type((u, v)))  # type: ignore (PyZX issue)
                elif add_cut_edge_as_boundary_node:
                    key = tuple(sorted((u, v)))
                    if key in cut_edges_map:
                        new_boundary_vertex = cut_edges_map[key]
                        input_vertices[new_boundary_vertex] = (v, u)
                    else:
                        new_boundary_vertex = -len(cut_edges_map) - 1
                        cut_edges_map[key] = new_boundary_vertex
                        output_vertices[new_boundary_vertex] = (v, u)
                    subgraph.add_vertex_indexed(new_boundary_vertex)
                    subgraph.set_type(new_boundary_vertex, VertexType.BOUNDARY)
                    subgraph.add_edge((v, new_boundary_vertex))
        subgraphs.append(subgraph)
        added_vertices_list.append((input_vertices, output_vertices))
    return subgraphs, added_vertices_list


def _partition_graph_into_connected_components(zx_graph: GraphS) -> list[GraphS]:
    """Partition the ZX graph into connected components."""
    visited = set()
    components = []
    for start_vertex in zx_graph.vertices():
        if start_vertex in visited:
            continue
        component_vertices = set()
        stack = [start_vertex]
        while stack:
            vertex = stack.pop()
            if vertex in visited:
                continue
            visited.add(vertex)
            component_vertices.add(vertex)
            stack.extend(
                neighbor for neighbor in zx_graph.neighbors(vertex) if neighbor not in visited
            )
        component = _partition_graph_from_vertices(zx_graph, [component_vertices], False)[0][0]
        components.append(component)
    return components


def _product_of_disconnected_correlation_surfaces(
    correlation_surfaces_list: Sequence[Sequence[_CorrelationSurfaceBase]],
) -> Iterator[_CorrelationSurfaceView]:
    """Generate correlation surfaces from the product of disconnected components."""
    return starmap(_CorrelationSurfaceView, product(*correlation_surfaces_list))


def _restore_correlation_surface_from_added_vertices(
    correlation_surface: _CorrelationSurfaceType,
    added_vertices: dict[int, tuple[int, int]],
) -> _CorrelationSurfaceType:
    """Restore the correlation surface by recovering the cut edges represented by boundary nodes."""
    for v, (u, w) in added_vertices.items():
        if u in correlation_surface and v in correlation_surface[u]:
            correlation_surface[u][w] = correlation_surface[u][v]
            del correlation_surface[u][v]
        if v in correlation_surface:
            del correlation_surface[v]
    return correlation_surface


def _find_correlation_surfaces_with_vertex_ordering(
    zx_graph: GraphS, vertex_ordering: Sequence[set[int]] | None = None, parallel: bool = False
) -> list[_CorrelationSurfaceView]:
    """Find the correlation surfaces based on a given vertex ordering."""
    if vertex_ordering is None:
        return list(
            _product_of_disconnected_correlation_surfaces(
                _find_correlation_surfaces(zx_graph, parallel)
            )
        )

    # partition the ZX graph and find correlation surface generators for each subgraph
    subgraphs, added_vertices_list = _partition_graph_from_vertices(zx_graph, vertex_ordering, True)
    if parallel and len(subgraphs) > 1:
        with multiprocessing.Pool() as pool:
            correlation_surfaces_list = pool.map(_find_correlation_surfaces, subgraphs)
    else:
        correlation_surfaces_list = list(map(_find_correlation_surfaces, subgraphs))

    # post-process the first subgraph
    out_vertices = added_vertices_list.pop(0)[1]
    valid_surfaces = list(
        _product_of_disconnected_correlation_surfaces(correlation_surfaces_list.pop(0))
    )
    stabilizers: list[dict[int, Pauli]] = [
        {
            v: p
            for v, p in zip(out_vertices, correlation_surface._paulis_at_nodes(out_vertices.keys()))
        }
        for correlation_surface in valid_surfaces
    ]  # the stabilizers need to be derived before the boundary vertices are restored
    valid_surfaces = [
        _restore_correlation_surface_from_added_vertices(g, out_vertices) for g in valid_surfaces
    ]

    # match the boundary stabilizers for the rest of the subgraphs
    for correlation_surfaces, (input_vertices, output_vertices) in zip(
        correlation_surfaces_list, added_vertices_list
    ):
        combinations = _product_of_disconnected_correlation_surfaces(correlation_surfaces)
        stabilizer_basis, basis_surfaces, invalid_stabilizer_indices = {}, [], []
        for i, stabilizer in enumerate(stabilizers):
            local_stabilizer = _concat_ints_as_bits(
                (
                    p.flipped(is_hadamard(zx_graph, input_vertices[v])).value
                    for v, p in stabilizer.items()
                    if v in input_vertices
                ),
                2,
            )
            for correlation_surface in combinations:
                _solve_linear_system(
                    stabilizer_basis, correlation_surface._signature_at_nodes(input_vertices.keys())
                )
                basis_surfaces.append(correlation_surface)
                indices = _solve_linear_system(
                    stabilizer_basis,
                    local_stabilizer,
                    False,
                )
                if indices is not None:
                    new_correlation_surface = _xor_correlation_surfaces(
                        [basis_surfaces[k] for k in indices]
                    )
                    stabilizer.update(
                        zip(
                            output_vertices,
                            new_correlation_surface._paulis_at_nodes(output_vertices.keys()),
                        )
                    )
                    for v in input_vertices.keys():
                        del stabilizer[v]
                    for vertices in (input_vertices, output_vertices):
                        _restore_correlation_surface_from_added_vertices(
                            new_correlation_surface, vertices
                        )
                    valid_surfaces[i].maps.append(new_correlation_surface)
                    break
            else:  # unsatisfiable stabilizer, remove it
                invalid_stabilizer_indices.append(i)
        stabilizers = [s for i, s in enumerate(stabilizers) if i not in invalid_stabilizer_indices]

    return valid_surfaces


def _find_correlation_surfaces(
    zx_graph: GraphS, parallel: bool = False
) -> list[list[_CorrelationSurface]]:
    """Find the correlation surface generators for each connected component in the graph."""
    components = [
        (component, min(v for v in component.vertices() if component.vertex_degree(v) == 1))
        for component in _partition_graph_into_connected_components(zx_graph)
    ]
    if parallel and len(components) > 1:
        with multiprocessing.Pool() as pool:
            correlation_surfaces = pool.starmap(_find_correlation_surfaces_from_leaf, components)
    else:
        correlation_surfaces = list(starmap(_find_correlation_surfaces_from_leaf, components))
    return correlation_surfaces


def _find_correlation_surfaces_from_leaf(zx_graph: GraphS, leaf: int) -> list[_CorrelationSurface]:
    """Find the correlation surface generators satisfying the closed ports."""
    correlation_surfaces = _find_correlation_surface_generating_set_from_leaf(zx_graph, leaf)

    leaves = {pauli: [] for pauli in Pauli.iter_ixzy()}
    for v in filter(
        lambda v: zx_graph.vertex_degree(v) == 1,
        zx_graph.vertices(),
    ):
        leaves[zx_to_pauli(zx_graph, v).flipped()].append(v)
    open_leaves = leaves.pop(Pauli.I)

    if sum(len(leaves) for leaves in leaves.values()):
        correlation_surfaces = _reform_correlation_surface_generators(
            correlation_surfaces,
            lambda cs: _concat_ints_as_bits(
                (
                    cs._signature_at_nodes(leaves, lambda p: p not in (Pauli.I, pauli), 1)
                    for pauli, leaves in leaves.items()
                ),
                map(len, leaves.values()),
            ),
            stabilizer_basis={},
            basis_surfaces=[],
        )[1]

    if open_leaves:
        correlation_surfaces = [
            _xor_correlation_surfaces([correlation_surfaces[i] for i in indices])
            if len(indices := _int_to_bit_indices(mask)) > 1
            else correlation_surfaces[indices[0]]  # type: ignore (indices is always non-empty)
            for _, mask in _normalize_basis(
                _construct_basis(
                    {},
                    correlation_surfaces,
                    lambda cs: cs._signature_at_nodes(open_leaves),
                )
            ).values()
        ]

    return correlation_surfaces  # ty: ignore (seems to be a false positive of ty)


def _reform_correlation_surface_generators(
    correlation_surfaces: Iterable[_CorrelationSurfaceType],
    signature_func: Callable[[_CorrelationSurfaceType], int],
    stabilizer_basis: dict[int, tuple[int, int]],
    basis_surfaces: Sequence[_CorrelationSurfaceType],
    construct_new_surfaces: bool = True,
    num_new_surfaces_needed: int | None = None,
    num_basis_surfaces_needed: int | None = None,
) -> tuple[list[_CorrelationSurfaceType], list[_CorrelationSurface]]:
    """Reform the correlation surface generators based on the given signature function."""
    basis_surfaces, new_surfaces = list(basis_surfaces), []
    for correlation_surface in correlation_surfaces:
        indices = _solve_linear_system(
            stabilizer_basis,
            signature_func(correlation_surface),
        )
        if indices is None:
            basis_surfaces.append(correlation_surface)
            if (
                num_basis_surfaces_needed is not None
                and len(basis_surfaces) >= num_basis_surfaces_needed
            ):
                break
            continue
        if construct_new_surfaces:
            new_surfaces.append(
                _xor_correlation_surfaces(
                    [*(basis_surfaces[k] for k in indices), correlation_surface]
                )
            )
            if num_new_surfaces_needed is not None and len(new_surfaces) >= num_new_surfaces_needed:
                break
    return basis_surfaces, new_surfaces


@cache
def _generate_valid_local_paulis(
    node_basis: Pauli,
    broadcast_pauli: Pauli,
    passthrough_parity: bool,
    num_unconnected_neighbors: int,
    generate_all: bool = False,
) -> list[list[Pauli]]:
    """Generate valid local Pauli configurations given broadcast and passthrough status."""
    unconnected_neighbors = range(num_unconnected_neighbors)
    combined_pauli = broadcast_pauli ^ node_basis
    if generate_all:
        passthrough_nodes = chain.from_iterable(
            combinations(unconnected_neighbors, num_passthrough)
            for num_passthrough in range(
                passthrough_parity,
                len(unconnected_neighbors) + 1,
                2,
            )
        )
        out_paulis_list = [
            [
                combined_pauli if n in passthrough_nodes else broadcast_pauli
                for n in unconnected_neighbors
            ]
            for passthrough_nodes in passthrough_nodes
        ]
    else:
        passthrough_nodes = (
            ((n,) for n in unconnected_neighbors)
            if passthrough_parity
            else pairwise(unconnected_neighbors)
        )
        out_paulis_list = [
            [combined_pauli if n in m else broadcast_pauli for n in unconnected_neighbors]
            for m in passthrough_nodes
        ]
        if not passthrough_parity or not num_unconnected_neighbors:
            out_paulis_list.append([broadcast_pauli] * num_unconnected_neighbors)
    return out_paulis_list


def _expand_correlation_surface_to_node(
    correlation_surface: _CorrelationSurface,
    broadcast_pauli: Pauli,
    passthrough_parity: bool,
    node: int,
    node_basis: Pauli,
    unconnected_neighbors: list[int],
    edges_are_hadamard: list[bool],
    generate_all: bool = True,
    always_copy: bool = False,
) -> Generator[_CorrelationSurface, None]:
    """Expand to a generating set of correlation surfaces for the new node."""
    for i, out_paulis in enumerate(
        _generate_valid_local_paulis(
            node_basis,
            broadcast_pauli,
            passthrough_parity,
            len(unconnected_neighbors),
            generate_all=generate_all,  # generate all can reduce the gaussian elimination overhead
        )
    ):
        if i or always_copy:
            new_correlation_surface = copy(correlation_surface)
            new_correlation_surface[node] = copy(correlation_surface[node])
        else:
            new_correlation_surface = correlation_surface
        for n, pauli, edge_is_hadamard in zip(
            unconnected_neighbors, out_paulis, edges_are_hadamard
        ):
            if (i or always_copy) and n in correlation_surface:
                new_correlation_surface[n] = copy(correlation_surface[n])
            new_correlation_surface._add_pauli_to_edge((node, n), pauli, edge_is_hadamard)
        yield new_correlation_surface


def _find_correlation_surface_generating_set_from_leaf(
    zx_graph: GraphS, leaf: int
) -> list[_CorrelationSurface]:
    """Find a generating set of correlation surfaces assuming all ports are open."""
    neighbor: int = next(iter(zx_graph.neighbors(leaf)))
    correlation_surfaces = (
        _CorrelationSurface()._add_pauli_to_edge(
            (leaf, neighbor), pauli, is_hadamard(zx_graph, (leaf, neighbor))
        )
        for pauli in Pauli.iter_xz()
    )
    if zx_graph.vertex_degree(neighbor) == 1:
        # edge case, make sure no leaf node will be in the frontier
        return list(correlation_surfaces)
    frontier = [neighbor]
    explored_leaves = [leaf]
    explored_nodes = {leaf}
    correlation_surface = next(correlation_surfaces)

    while frontier:
        current_node = frontier.pop(0)
        connected_neighbors = list(correlation_surface[current_node].keys())
        unconnected_neighbors = list(
            filter(
                lambda n: n not in connected_neighbors,
                zx_graph.neighbors(current_node),
            )
        )
        boundary_nodes = explored_leaves + frontier
        if unconnected_neighbors:
            boundary_nodes.append(current_node)
        generating_set_size = sum(len(correlation_surface[n]) for n in boundary_nodes)
        unexplored_neighbors = [n for n in unconnected_neighbors if n not in correlation_surface]
        passthrough_basis = zx_to_pauli(zx_graph, current_node)

        # check if each correlation surface candidate satisfies broadcast and passthrough rules
        # on the current node and is not a product of previously checked valid correlation surfaces
        valid_surfaces, invalid_surfaces, syndromes, vector_basis = [], [], [], {}
        for correlation_surface in chain([correlation_surface], correlation_surfaces):
            constraint_check = correlation_surface._validate_node(
                current_node, passthrough_basis, bool(unconnected_neighbors)
            )
            if isinstance(constraint_check, int):  # invalid
                invalid_surfaces.append(correlation_surface)
                syndromes.append(constraint_check)
                continue
            if (
                _solve_linear_system(
                    vector_basis, correlation_surface._signature_at_nodes(boundary_nodes)
                )
                is None
            ):  # new independent surface
                valid_surfaces.append((correlation_surface, *constraint_check))
                if len(vector_basis) == generating_set_size:
                    break

        # try to fix local constraint violations by XORing with other invalid surfaces
        all_one = (1 << len(connected_neighbors)) - 1
        syndrome_basis, basis_surfaces = {}, []
        for correlation_surface, syndrome in zip(invalid_surfaces, syndromes):
            if len(vector_basis) == generating_set_size:
                break
            for j, target in enumerate((syndrome ^ all_one, syndrome)):  # two valid options
                indices = _solve_linear_system(syndrome_basis, target, update_basis=j == 1)
                if indices is None:
                    if j == 1:
                        basis_surfaces.append(correlation_surface)
                    continue
                new_correlation_surface = _xor_correlation_surfaces(
                    [*(basis_surfaces[k] for k in indices), correlation_surface]
                )
                if (
                    _solve_linear_system(
                        vector_basis, new_correlation_surface._signature_at_nodes(boundary_nodes)
                    )
                    is None
                ):
                    valid_surfaces.append(
                        (
                            new_correlation_surface,
                            *new_correlation_surface._validate_node(
                                current_node, passthrough_basis, bool(unconnected_neighbors)
                            ),  # type: ignore (the new surface is always valid so it's always a tuple)
                        )
                    )
                    break

        # enumerate new branches
        correlation_surfaces = chain.from_iterable(
            starmap(
                partial(
                    _expand_correlation_surface_to_node,
                    node=current_node,
                    node_basis=passthrough_basis,
                    unconnected_neighbors=unconnected_neighbors,
                    edges_are_hadamard=[
                        is_hadamard(zx_graph, (current_node, n)) for n in unconnected_neighbors
                    ],
                    # always_copy=True,  # can be False if the exploration is BFS
                ),
                valid_surfaces,
            )
        )
        correlation_surface = next(correlation_surfaces, None)
        if not correlation_surface:  # no valid correlation surface exists on this ZX graph
            return []
        frontier.extend(
            filter(
                lambda n: n not in explored_nodes and zx_graph.vertex_degree(n) > 1,
                unexplored_neighbors,
            )
        )
        explored_leaves.extend(
            filter(lambda n: zx_graph.vertex_degree(n) == 1, unexplored_neighbors)
        )
        explored_nodes.add(current_node)

    # eliminate dependent surfaces from the final expansion to get a generating set
    return _reform_correlation_surface_generators(
        [correlation_surface, *correlation_surfaces],
        lambda cs: cs._signature_at_nodes(
            filter(
                lambda v: zx_graph.vertex_degree(v) == 1,
                zx_graph.vertices(),
            )
        ),
        stabilizer_basis={},
        basis_surfaces=[],
        construct_new_surfaces=False,
        num_basis_surfaces_needed=len(
            [v for v in zx_graph.vertices() if zx_graph.vertex_degree(v) == 1]
        ),
    )[0]


def _concat_ints_as_bits(ints: Iterable[int], bit_length: int | Iterable[int]) -> int:
    """Concatenate a list of integers as bits to form a single integer."""
    if isinstance(bit_length, int):
        bit_length = repeat(bit_length)
    return sum(x << shift for x, shift in zip(ints, chain([0], accumulate(bit_length))))


def _solve_linear_system(
    basis: dict[int, tuple[int, int]], x: int, update_basis: bool = True
) -> tuple[int, ...] | None:
    """Gaussian elimination over GF(2)."""
    mask = 1 << len(basis)
    while x:
        highest_bit = x.bit_length() - 1
        if highest_bit not in basis:
            if update_basis:
                basis[highest_bit] = (x, mask)
            return
        pivot, pivot_mask = basis[highest_bit]
        x ^= pivot
        mask ^= pivot_mask
    return _int_to_bit_indices(mask)[:-1]


def _int_to_bit_indices(x: int) -> tuple[int, ...]:
    """Convert an integer to a list of indices where the bits are set."""
    return tuple(i for i in range(x.bit_length()) if (x >> i) & 1)


def _construct_basis(
    basis: dict[int, tuple[int, int]], items: Iterable[Any], func: Callable[[Any], int]
) -> dict[int, tuple[int, int]]:
    """Construct a linear basis from the given items using the provided function."""
    for item in items:
        _solve_linear_system(basis, func(item))
    return basis


def _normalize_basis(
    basis: dict[int, tuple[int, int]], in_place: bool = True
) -> dict[int, tuple[int, int]]:
    """Normalize the basis vectors to only have leading 1s when possible."""
    normalized_basis = basis if in_place else {}
    highest_bits = sorted(basis, reverse=True)
    for i, key in enumerate(highest_bits):
        vector, mask = basis[key]
        for highest_bit in highest_bits[i + 1 :]:
            if (vector >> highest_bit) & 1:
                pivot, pivot_mask = basis[highest_bit]
                vector ^= pivot
                mask ^= pivot_mask
        normalized_basis[key] = (vector, mask)
    return normalized_basis


def _check_spiders_are_supported(g: GraphS) -> None:
    """Check the preconditions for the correlation surface finding algorithm."""
    for v in g.vertices():
        # 1. Check the spider types and phases are supported
        vt = g.type(v)
        phase = g.phase(v)
        try:
            pauli = vertex_type_to_pauli(vt, phase)
        except TQECError:
            raise TQECError(f"Unsupported spider type and phase: {vt} and {phase}.")
        # 2. Check degree of the spiders
        degree = g.vertex_degree(v)
        if pauli is Pauli.I and degree != 1:
            raise TQECError(f"Boundary spider must be dangling, but got {degree} neighbors.")
        if pauli is Pauli.Y and degree != 1:
            raise TQECError(f"S spider must be dangling, but got {degree} neighbors.")



"""Defines core data-structures to handle the mapping between qubit coordinates and qubit indices.

A bijection from qubit coordinates to qubit indices is represented by
:class:`QubitMap` defined in this module.

This bijection can be obtained from a ``stim.Circuit`` instance by using
:func:`get_qubit_map`.

"""

from __future__ import annotations

import functools
from collections import Counter
from collections.abc import Callable, Iterable
from dataclasses import dataclass, field
from typing import Any

import numpy
import stim

from tqec.circuit.qubit import GridQubit
from tqec.utils.coordinates import StimCoordinates
from tqec.utils.exceptions import TQECError
from tqec.utils.scale import round_or_fail


@dataclass(frozen=True)
class QubitMap:
    """Represent a bijection between :class:`~tqec.circuit.qubit.GridQubit` instances and indices.

    This class aims at representing a bidirectional mapping (hence the
    "bijection") between qubits and their associated indices.

    Raises:
        TQECError: if the provided mapping from indices to qubits is not a
            bijection (i.e., if at least to values represent the same qubit).

    """

    i2q: dict[int, GridQubit] = field(default_factory=dict)

    def __post_init__(self) -> None:
        qubit_counter = Counter(self.i2q.values())
        if len(qubit_counter) != len(self.i2q):
            duplicated_qubits = frozenset(q for q in qubit_counter if qubit_counter[q] > 1)
            raise TQECError(f"Found qubit(s) with more than one index: {duplicated_qubits}.")

    @staticmethod
    def from_qubits(qubits: Iterable[GridQubit]) -> QubitMap:
        """Create a qubit map from the provided ``qubits``.

        Qubit indices are associated in the order in which ``qubits`` are provided: the first qubit
        will have index ``0``, the second index ``1`` and so on.
        """
        return QubitMap(dict(enumerate(qubits)))

    @staticmethod
    def from_circuit(circuit: stim.Circuit) -> QubitMap:
        """Return a qubit map from the qubit coordinates at the end of the provided ``circuit``.

        Warning:
            This function, just like
            `stim.Circuit.get_final_qubit_coordinates <https://github.com/quantumlib/Stim/blob/main/doc/python_api_reference_vDev.md#stim.Circuit.get_final_qubit_coordinates>`_,
            returns the qubit coordinates **at the end** of the provided ``circuit``.

        Args:
            circuit: instance to get qubit coordinates from.

        Raises:
            TQECError: if any of the final qubits is not defined with exactly 2
                coordinates (we only consider qubits on a 2-dimensional grid).

        Returns:
            a mapping from qubit indices (keys) to qubit coordinates (values).

        """
        return get_qubit_map(circuit)

    @functools.cached_property
    def q2i(self) -> dict[GridQubit, int]:
        """Get a mapping from qubits to indices."""
        return {q: i for i, q in self.i2q.items()}

    @property
    def indices(self) -> Iterable[int]:
        """Get all the qubit indices managed by ``self``."""
        return self.i2q.keys()

    @property
    def qubits(self) -> Iterable[GridQubit]:
        """Get all the qubits manager by ``self``."""
        return self.i2q.values()

    def with_mapped_qubits(self, qubit_map: Callable[[GridQubit], GridQubit]) -> QubitMap:
        """Change the qubits involved in ``self`` without changing the associated indices.

        Args:
            qubit_map: a map from qubits to qubits that should associate a qubit
                to each of the qubits represented by ``self``.

        Raises:
            KeyError: if any qubit in ``self.qubits`` is not present in the keys of
                the provided ``qubit_map``.

        Returns:
            a new instance representing the updated mapping.

        """
        return QubitMap({i: qubit_map(q) for i, q in self.i2q.items()})

    def items(self) -> Iterable[tuple[int, GridQubit]]:
        """Get an iterator over each qubit and its corresponding index."""
        return self.i2q.items()

    def filter_by_qubits(self, qubits_to_keep: Iterable[GridQubit]) -> QubitMap:
        """Filter the qubit map to only keep qubits present in the provided ``qubits_to_keep``.

        Args:
            qubits_to_keep: the qubits to keep in the circuit.

        Returns:
            a copy of ``self`` for which the assertion
            ``set(return_value.qubits).issubset(qubits_to_keep)`` is ``True``.

        """
        return self.filter_by_qubit_indices(self.q2i[q] for q in qubits_to_keep if q in self.q2i)

    def filter_by_qubit_indices(self, qubit_indices_to_keep: Iterable[int]) -> QubitMap:
        """Filter the qubit map to only keep qubits present in ``qubit_indices_to_keep``.

        Args:
            qubit_indices_to_keep: the qubits to keep in the circuit.

        Returns:
            a copy of ``self`` with all the qubits associated to indices not in
            ``qubit_indices_to_keep`` removed.

        """
        kept_qubit_indices = frozenset(qubit_indices_to_keep)
        return QubitMap({i: q for i, q in self.i2q.items() if i in kept_qubit_indices})

    def to_circuit(self, shift_to_positive: bool = False) -> stim.Circuit:
        """Get a circuit with only ``QUBIT_COORDS`` instructions representing ``self``.

        Args:
            shift_to_positive: if ``True``, the qubit coordinates are shift such
                that they are all positive. Their relative positioning stays
                unchanged.

        Returns:
            a ``stim.Circuit`` containing only ``QUBIT_COORDS`` instructions.

        """
        shiftx, shifty = 0, 0
        if shift_to_positive:
            shiftx = -min((q.x for q in self.i2q.values()), default=0)
            shifty = -min((q.y for q in self.i2q.values()), default=0)
        ret = stim.Circuit()
        for qi, qubit in sorted(self.i2q.items(), key=lambda t: t[0]):
            ret.append(
                "QUBIT_COORDS",
                qi,
                StimCoordinates(qubit.x + shiftx, qubit.y + shifty).to_stim_coordinates(),
            )
        return ret

    def __getitem__(self, index: GridQubit) -> int:
        return self.q2i[index]

    def bounding_box(self) -> tuple[tuple[float, float], tuple[float, float]]:
        """Compute and return the bounding box of ``self``.

        Returns:
            ``(mins, maxes)`` where each of ``mins`` (resp. ``maxes``) is a pair
            of values ``(x, y)`` corresponding to the dimension.

        """
        coordinates = numpy.array([(q.x, q.y) for q in self.qubits])
        mins = numpy.min(coordinates, axis=0)
        maxes = numpy.max(coordinates, axis=0)
        return ((mins[0], mins[1]), (maxes[0], maxes[1]))

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the qubit map.

        The dictionary is intended to be used as a JSON object.

        """
        return {"i2q": [[qi, q.to_dict()] for qi, q in self.i2q.items()]}

    @staticmethod
    def from_dict(data: dict[str, Any]) -> QubitMap:
        """Return a qubit map from its dictionary representation.

        Args:
            data: dictionary with the keys ``i2q`` and ``q2i``.

        Returns:
            a new instance of :class:`QubitMap` with the provided
            ``i2q`` and ``q2i``.

        """
        i2q = {int(qi): GridQubit.from_dict(q) for qi, q in data["i2q"]}
        return QubitMap(i2q)

    def qubit_bounds(self) -> tuple[GridQubit, GridQubit]:
        """Return the tightest possible bounding box containing all the qubits in ``self``.

        Raises:
            TQECError: if ``self`` is empty.

        Returns:
            ``(top_left, bottom_right)`` representing the bounding box of the qubits listed in
            ``self``.

        """
        if not self.i2q:
            raise TQECError("Cannot get the bounding box of an empty QubitMap.")
        qxs, qys = [q.x for q in self.i2q.values()], [q.y for q in self.i2q.values()]
        return GridQubit(min(qxs), min(qys)), GridQubit(max(qxs), max(qys))


def get_qubit_map(circuit: stim.Circuit) -> QubitMap:
    """Return the existing qubits and their coordinates at the end of the provided ``circuit``.

    Warning:
        This function, just like
        `stim.Circuit.get_final_qubit_coordinates <https://github.com/quantumlib/Stim/blob/main/doc/python_api_reference_vDev.md#stim.Circuit.get_final_qubit_coordinates>`_,
        returns the qubit coordinates **at the end** of the provided ``circuit``.

    Args:
        circuit: instance to get qubit coordinates from.

    Raises:
        TQECError: if any of the final qubits is not defined with exactly 2
            coordinates (we only consider qubits on a 2-dimensional grid).

    Returns:
        a mapping from qubit indices (keys) to qubit coordinates (values).

    """
    qubit_coordinates = circuit.get_final_qubit_coordinates()
    qubits: dict[int, GridQubit] = {}
    for qi, coords in qubit_coordinates.items():
        if len(coords) != 2:
            raise TQECError(
                "Qubits should be defined on exactly 2 spatial dimensions. "
                f"Found {qi} -> {coords} defined on {len(coords)} spatial dimensions."
            )
        x = round_or_fail(coords[0])
        y = round_or_fail(coords[1])
        qubits[qi] = GridQubit(x, y)
    return QubitMap(qubits)



"""Utility functions for PyZX interop."""

from fractions import Fraction

from pyzx.graph.graph_s import GraphS
from pyzx.utils import EdgeType, FractionLike, VertexType, vertex_is_zx

from tqec.computation.cube import CubeKind, Port, ZXCube
from tqec.utils.enums import Basis, Pauli
from tqec.utils.exceptions import TQECError


def is_zx_no_phase(g: GraphS, v: int) -> bool:
    """Check if a vertex in a PyZX graph is a Z/X spider with phase 0."""
    return vertex_is_zx(g.type(v)) and g.phase(v) == 0


def is_z_no_phase(g: GraphS, v: int) -> bool:
    """Check if a vertex in a PyZX graph is a Z spider with phase 0."""
    return g.type(v) is VertexType.Z and g.phase(v) == 0


def is_x_no_phase(g: GraphS, v: int) -> bool:
    """Check if a vertex in a PyZX graph is a X spider with phase 0."""
    return g.type(v) is VertexType.X and g.phase(v) == 0


def is_boundary(g: GraphS, v: int) -> bool:
    """Check if a vertex in a PyZX graph is a boundary type spider."""
    return g.type(v) is VertexType.BOUNDARY


def is_s(g: GraphS, v: int) -> bool:
    """Check if a vertex in a PyZX graph is a S node."""
    return g.type(v) is VertexType.Z and g.phase(v) == Fraction(1, 2)


def is_hadamard(g: GraphS, edge: tuple[int, int]) -> bool:
    """Check if an edge in a PyZX graph is a Hadamard edge."""
    return g.edge_type(edge) is EdgeType.HADAMARD


def cube_kind_to_zx(kind: CubeKind) -> tuple[VertexType, FractionLike]:
    """Convert the cube kind to the corresponding PyZX vertex type and phase.

    The conversion is as follows:

    - Port -> BOUNDARY spider with phase 0.
    - YHalfCube -> Z spider with phase 1/2.
    - ZXCube -> Z spider with phase 0 if it has only one Z basis boundary,
        otherwise X spider with phase 0.

    Args:
        kind: The cube kind to be converted.

    Returns:
        A tuple of vertex type and spider phase.

    """
    if isinstance(kind, ZXCube):
        if sum(basis == Basis.Z for basis in kind.as_tuple()) == 1:
            return VertexType.Z, 0
        return VertexType.X, 0
    if isinstance(kind, Port):
        return VertexType.BOUNDARY, 0
    else:  # isinstance(kind, YHalfCube)
        return VertexType.Z, Fraction(1, 2)


def zx_to_pauli(g: GraphS, v: int) -> Pauli:
    """Convert a PyZX vertex to the corresponding Pauli operator.

    Args:
        g: The PyZX graph.
        v: The vertex id.

    Raises:
        ValueError: If the vertex is not a Clifford or a boundary.

    Returns:
        The corresponding Pauli operator.

    """
    return vertex_type_to_pauli(g.type(v), g.phase(v))


def vertex_type_to_pauli(vertex_type: VertexType, phase: FractionLike = 0) -> Pauli:
    """Convert a PyZX vertex type to the corresponding Pauli operator.

    Args:
        vertex_type: The PyZX vertex type.
        phase: The phase of the vertex. Default is 0.

    Raises:
        TQECError: If the vertex type and phase do not correspond to a Pauli operator.

    Returns:
        The corresponding Pauli operator.

    """
    match vertex_type, phase:
        case VertexType.X, 0:
            return Pauli.X
        case VertexType.Z, 0:
            return Pauli.Z
        case VertexType.Z, Fraction(numerator=1, denominator=2):
            return Pauli.Y
        case VertexType.BOUNDARY, _:
            return Pauli.I
        case _:
            raise TQECError(
                f"Cannot convert vertex type {vertex_type} and phase {phase} to Pauli operator."
            )


def zx_to_basis(g: GraphS, v: int) -> Basis:
    """Convert a PyZX vertex to the corresponding Basis.

    Args:
        g: The PyZX graph.
        v: The vertex id.

    Raises:
        ValueError: If the vertex is not a Clifford or a boundary.

    Returns:
        The corresponding Basis.

    """
    return vertex_type_to_basis(g.type(v), g.phase(v))


def vertex_type_to_basis(vertex_type: VertexType, phase: FractionLike = 0) -> Basis:
    """Convert a PyZX vertex type to the corresponding Basis.

    Args:
        vertex_type: The PyZX vertex type.
        phase: The phase of the vertex. Default is 0.

    Raises:
        TQECError: If the vertex type and phase do not correspond to a Basis.

    Returns:
        The corresponding Basis.

    """
    try:
        return vertex_type_to_pauli(vertex_type, phase).to_basis()
    except TQECError:
        raise TQECError(f"Cannot convert vertex type {vertex_type} and phase {phase} to Basis.")



import math
from abc import ABC, abstractmethod
from enum import Enum
from typing import ClassVar, Final

import svg
from typing_extensions import override

from tqec.interop.color import TQECColor
from tqec.plaquette.enums import PlaquetteOrientation
from tqec.plaquette.rpng.rpng import ExtendedBasis, PauliBasis
from tqec.utils.enums import Basis
from tqec.visualisation.configuration import DrawerConfiguration


def lerp(a: complex, b: complex, t: float) -> complex:
    """Linear interpolation between ``a`` and ``b``.

    Returns:
        a linear interpolation between ``a`` and ``b`` according to the provided
        ``t``: ``(1 - t) * a + t * b``.

    """
    return (1 - t) * a + t * b


class PlaquetteCorner(Enum):
    TOP_LEFT = "TL"
    TOP_RIGHT = "TR"
    BOTTOM_LEFT = "BL"
    BOTTOM_RIGHT = "BR"


class SVGPlaquetteDrawer(ABC):
    """Base class for plaquette drawers that output to SVG.

    A few static helper methods are defined in this class and can be re-used by sub-classes, for
    example to get a SVG path from a set of points.

    """

    _CENTER_COORDINATE: ClassVar[complex] = 0.5 + 0.5j
    _CORNERS: ClassVar[list[complex]] = [0, 1, 1.0j, 1 + 1.0j]
    _CORNERS_ENUM: ClassVar[list[PlaquetteCorner]] = [
        PlaquetteCorner.TOP_LEFT,
        PlaquetteCorner.TOP_RIGHT,
        PlaquetteCorner.BOTTOM_LEFT,
        PlaquetteCorner.BOTTOM_RIGHT,
    ]

    @abstractmethod
    def draw(
        self,
        id: str,
        show_interaction_order: bool = True,
        show_hook_errors: bool = True,
        show_data_qubit_reset_measurements: bool = True,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Element:
        """Draw the plaquette and return the drawing as a SVG element.

        Args:
            id: a unique ID representing the drawing. Used internally to reference the drawing. A
                good default value could be the name of the drawn plaquette.
            show_interaction_order: if ``True``, the time steps at which data-qubits interact with
                the syndrome qubit(s) is written on the drawing.
            show_hook_errors: if ``True``, a line is drawn to represent the hook error when it
                exists.
            show_data_qubit_reset_measurements: if ``True``, small symbols are used on data-qubits
                when a reset or a measurement is applied.
            configuration: drawing configuration.

        Returns:
            the drawing as a SVG element.

        """
        pass

    @staticmethod
    def get_square_shape(
        fill: str = "none",
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Path:
        """Get the shape of a regular square plaquette."""
        return svg_path_enclosing_points(SVGPlaquetteDrawer._CORNERS, fill, configuration)

    @staticmethod
    def get_triangle_shape(
        missing_qubit_corner: PlaquetteCorner,
        fill: str = "none",
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.G:
        """Get the shape of a triangular plaquette only involving 3 data qubits.

        Args:
            missing_qubit_corner: the plaquette corner the missing qubit is located on.
            fill: SVG colour to use to fill the triangle.
            configuration: drawing configuration.

        """
        # Draw a TOP_LEFT corner 3-qubit triangular plaquette by default.
        _corners: Final[list[complex]] = SVGPlaquetteDrawer._CORNERS
        default_corners = [_corners[i] for i in [1, 2, 3]]
        t = configuration.plaquette_overflow_lerp_coefficient
        shoulders = [lerp(_corners[1], _corners[0], t), lerp(_corners[2], _corners[0], t)]
        default_triangular_path = svg_path_enclosing_points(
            default_corners + shoulders, fill, configuration
        )
        # Rotate the drawn plaquette if needed.
        rotation_angle: int = {
            PlaquetteCorner.TOP_LEFT: 0,
            PlaquetteCorner.TOP_RIGHT: 90,
            PlaquetteCorner.BOTTOM_LEFT: 270,
            PlaquetteCorner.BOTTOM_RIGHT: 180,
        }[missing_qubit_corner]
        return svg.G(
            transform=[
                svg.Rotate(
                    rotation_angle,
                    SVGPlaquetteDrawer._CENTER_COORDINATE.real,
                    SVGPlaquetteDrawer._CENTER_COORDINATE.imag,
                )
            ],
            elements=[default_triangular_path],
        )

    @staticmethod
    def get_half_circle_shape(
        orientation: PlaquetteOrientation,
        fill: str = "none",
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Path:
        """Get the shape of a half-circle plaquette only involving 2 data qubits.

        Args:
            orientation: half circle plaquette orientation. ``UP`` means that the circular part is
                pointing up.
            fill: SVG colour to use to fill the square.
            configuration: drawing configuration.

        """
        match orientation:
            case PlaquetteOrientation.UP:
                start, end, sweep = (0, 1), (1, 1), True
            case PlaquetteOrientation.DOWN:
                start, end, sweep = (0, 0), (1, 0), False
            case PlaquetteOrientation.LEFT:
                start, end, sweep = (1, 0), (1, 1), False
            case _:
                start, end, sweep = (0, 0), (0, 1), True

        return svg.Path(
            d=[
                svg.M(*start),
                svg.Arc(0.5, 0.5, 180, True, sweep, *end),
                svg.Z(),
            ],
            fill=fill,
            stroke=configuration.stroke_color,
            stroke_width=configuration.stroke_width,
        )

    @staticmethod
    def get_reset_shape(
        place: PlaquetteCorner,
        fill: str = "none",
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Rect:
        """Return a small square with its origin at the opposite of ``place``.

        Args:
            place: the plaquette corner the reset is located on.
            fill: SVG colour to use to fill the square.
            configuration: drawing configuration.

        Returns:
            the reset shape (a small square) that can then be applied the appropriate offset to be
            placed in the ``place`` corner of the plaquette.

        """
        r = configuration.reset_square_radius
        return svg.Rect(
            x=0 if place in [PlaquetteCorner.TOP_LEFT, PlaquetteCorner.BOTTOM_LEFT] else -r,
            y=0 if place in [PlaquetteCorner.TOP_LEFT, PlaquetteCorner.TOP_RIGHT] else -r,
            width=r,
            height=r,
            fill=fill,
            stroke=configuration.stroke_color,
            stroke_width=configuration.stroke_width,
        )

    @staticmethod
    def get_measurement_shape(
        place: PlaquetteCorner,
        fill: str = "none",
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.G:
        """Return a small quarter circle with its origin at the opposite of ``place``.

        Args:
            place: the plaquette corner the measurement is located on.
            fill: SVG colour to use to fill the quarter circle.
            configuration: drawing configuration.

        Returns:
            the measurement shape (a small quarter circle) that can then be applied the appropriate
            offset to be placed in the ``place`` corner of the plaquette.

        """
        rotation_angle: int = {
            PlaquetteCorner.TOP_LEFT: 0,
            PlaquetteCorner.BOTTOM_LEFT: 270,
            PlaquetteCorner.TOP_RIGHT: 90,
            PlaquetteCorner.BOTTOM_RIGHT: 180,
        }[place]
        r = configuration.measurement_circle_radius
        return svg.G(
            elements=[
                svg.Path(
                    d=[
                        svg.M(0, 0),
                        svg.L(0, r),
                        svg.Arc(r, r, 90, False, False, r, 0),
                        svg.Z(),
                    ],
                    stroke=configuration.stroke_color,
                    stroke_width=configuration.stroke_width,
                    fill=fill,
                )
            ],
            transform=[svg.Rotate(rotation_angle, 0, 0)],
        )

    @staticmethod
    def get_colour(basis: Basis | PauliBasis | ExtendedBasis) -> str:
        """Get an SVG-compatible hexadecimal color from a basis."""
        match basis.value.upper():
            case "X":
                return TQECColor.X.rgba.to_hex()
            case "Y":
                return TQECColor.Y.rgba.to_hex()
            case "Z":
                return TQECColor.Z.rgba.to_hex()
            case _:
                return "none"

    @staticmethod
    def get_corner_coordinates(corner: PlaquetteCorner) -> complex:
        """Get the coordinates of the point corresponding the provided ``corner``."""
        match corner:
            case PlaquetteCorner.TOP_LEFT:
                return SVGPlaquetteDrawer._CORNERS[0]
            case PlaquetteCorner.TOP_RIGHT:
                return SVGPlaquetteDrawer._CORNERS[1]
            case PlaquetteCorner.BOTTOM_LEFT:
                return SVGPlaquetteDrawer._CORNERS[2]
            case PlaquetteCorner.BOTTOM_RIGHT:
                return SVGPlaquetteDrawer._CORNERS[3]


class EmptySVGPlaquetteDrawer(SVGPlaquetteDrawer):
    """SVG plaquete drawer that always returns an empty drawing."""

    @override
    def draw(
        self,
        id: str,
        show_interaction_order: bool = True,
        show_hook_errors: bool = True,
        show_data_qubit_reset_measurements: bool = True,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Element:
        return svg.G()


def _sort_by_angle(center: complex, points: list[complex]) -> list[complex]:
    """Sort the given ``points`` w.r.t to the angle they form with the provided ``center``."""
    translated_points = [p - center for p in points]
    sorted_translated_points = sorted(translated_points, key=lambda c: math.atan2(c.imag, c.real))
    return [tp + center for tp in sorted_translated_points]


def svg_path_enclosing_points(
    points: list[complex],
    fill: str = "none",
    configuration: DrawerConfiguration = DrawerConfiguration(),
) -> svg.Path:
    """Draw a path enclosing all the provided points.

    Args:
        points: a list of points that will be linked together.
        fill: hexadecimal colour used to fill the produced path.
        configuration: drawing configuration.

    Warning:
        This function does **NOT** returns the convex hull of the provided points
        as a SVP path. It simply links the provided points one after the other
        after ordering them according to the angle they form with respect to their
        average (equivalent to their center of mass where each point has a mass
        of ``1``).

    Returns:
        a closed path enclosing all the provided ``points``.

    """
    center_point: complex = sum(points) / len(points)
    first, *others = _sort_by_angle(center_point, points)
    pathdata: list[svg.PathData] = [svg.M(first.real, first.imag)]
    pathdata.extend(svg.L(p.real, p.imag) for p in others)
    pathdata.append(svg.Z())
    return svg.Path(
        d=pathdata,
        fill=fill,
        stroke=configuration.stroke_color,
        stroke_width=configuration.stroke_width,
    )



"""Defines the color representation used in the TQEC interop layer."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum


@dataclass(frozen=True)
class RGBA:
    """RGBA representation of a color.

    Attributes:
        r: Red component of the color, in the range [0, 255].
        g: Green component of the color, in the range [0, 255].
        b: Blue component of the color, in the range [0, 255].
        a: Alpha component of the color, in the range [0, 1].

    """

    r: int
    g: int
    b: int
    a: float

    def with_alpha(self, a: float) -> RGBA:
        """Return a new RGBA with the same color but a different alpha value.

        Args:
            a: The new alpha value.

        Returns:
            A new RGBA with the same color but a different alpha value.

        """
        return RGBA(self.r, self.g, self.b, a)  # pragma: no cover

    def as_floats(self) -> tuple[float, float, float, float]:
        """Return the color as a tuple of floats. RGB values are normalized to the range [0, 1].

        Returns:
            Length 4 tuple of floats representing the color.

        """
        return (self.r / 255, self.g / 255, self.b / 255, self.a)

    def to_hex(self) -> str:
        """Return the color as a hexadecimal string.

        Returns:
            The color as a hexadecimal string.

        """
        return f"#{self.r:02x}{self.g:02x}{self.b:02x}"  # pragma: no cover


class TQECColor(Enum):
    """Predefined and commonly used colors in TQEC."""

    X = "X"
    Y = "Y"
    Z = "Z"
    H = "H"
    X_CORRELATION = "X_CORRELATION"
    Z_CORRELATION = "Z_CORRELATION"

    @property
    def rgba(self) -> RGBA:
        """Return the RGBA representation of the color."""
        if self == TQECColor.X:
            return RGBA(255, 127, 127, 1.0)
        if self == TQECColor.Y:
            return RGBA(99, 198, 118, 1.0)
        if self == TQECColor.Z:
            return RGBA(115, 150, 255, 1.0)
        if self == TQECColor.H:
            return RGBA(255, 255, 101, 1.0)
        if self == TQECColor.X_CORRELATION:
            return RGBA(255, 0, 0, 0.8)
        else:  # if self == TQECColor.Z_CORRELATION:
            return RGBA(0, 0, 255, 0.8)

    def with_zx_flipped(self) -> TQECColor:
        """Return a ``X`` or ``Z`` color from a ``Z`` or ``X`` color and vice versa."""
        if self == TQECColor.X:
            return TQECColor.Z
        if self == TQECColor.Z:
            return TQECColor.X
        if self == TQECColor.X_CORRELATION:
            return TQECColor.Z_CORRELATION
        if self == TQECColor.Z_CORRELATION:
            return TQECColor.X_CORRELATION
        return self



"""Split the statistics for multiple observables."""

import collections
from collections.abc import Mapping

import sinter

# Import of a private module not marked as explicitly typed, type ignore for mypy.
from sinter._data import ExistingData  # type: ignore

from tqec.computation.correlation import CorrelationSurface


def split_counts_for_observables(counts: Mapping[str, int], num_observables: int) -> list[int]:
    """Split the error counts for each individual observable.

    This function should only be used when specifying ``count_observable_error_combos=True`` to
    ``sinter`` functions.

    Args:
        counts: The error counts for different observable error combinations.
        num_observables: The number of observables.

    Returns:
        A list of error counts for each individual observable.

    """
    split_counts: list[int] = [0] * num_observables
    for key, count in counts.items():
        if not key.startswith("obs_mistake_mask="):
            continue
        comb = key.split("=")[1]
        assert num_observables == len(comb)
        for i in range(num_observables):
            if comb[i] == "E":
                split_counts[i] += count
    return split_counts


def split_stats_for_observables(
    stats: list[sinter.TaskStats],
    num_observables: int,
) -> list[list[sinter.TaskStats]]:
    """Split the statistics for each individual observable.

    This function should only be used when specifying ``count_observable_error_combos=True`` to
    ``sinter`` functions.

    Args:
        stats: The statistics for different observable error combinations.
        num_observables: number of observables contained in the provided ``stats``.

    Returns:
        A list of statistics for each individual observable.

    """
    # Combine the stats for each task
    data = ExistingData()
    for s in stats:
        data.add_sample(s)
    combined_stats = list(data.data.values())

    # For each task, split the stats by observable
    stats_by_observables: list[list[sinter.TaskStats]] = [[] for _ in range(num_observables)]
    for task_stats in combined_stats:
        split_counts = split_counts_for_observables(task_stats.custom_counts, num_observables)
        for obs_idx, count in enumerate(split_counts):
            stats_by_observables[obs_idx].append(
                task_stats.with_edits(
                    errors=count,
                    custom_counts=collections.Counter(
                        {
                            k: v
                            for k, v in task_stats.custom_counts.items()
                            if not k.startswith("obs_mistake_mask=")
                        }
                    ),
                )
            )
    return stats_by_observables


def heuristic_custom_error_key(observables: list[CorrelationSurface]) -> str:
    """Get a heuristic custom error key using for the ``error`` metric in ``sinter``.

    To get the statistics for each individual observable, we set
    ``count_observable_error_combos=True`` in ``sinter.collect``. It will include
    error counts for each observable error combination in the stats. For example,
    ``{'obs_mistake_mask=EE_': 10, 'obs_mistake_mask=E_E': 20}`` means that 10
    errors flipped both the first and second observables, and 20 errors flipped
    both the first and third observables, are sampled. We then split the counts
    to get the error counts for each individual observable.

    By default, ``sinter`` will count an error that flips any number of observables
    as a single error in the stats. As we want to sample enough statistics for
    each individual observable, which might have large variance in the logical
    error rate, we need to select a good ``custom_error_count_key``.

    Firstly we find the observable with the minimum error rate, which is expected
    to be the observable with the minimal correlation surface area. If we ensure
    this observable has enough statistical data, we can ensure all the observables
    have enough data. Therefore, we need to find all the observable combinations
    that contains this observable. Then we order the combinations by their
    probabilities, and return the most frequent one as the heuristic custom error
    key, which needs the least shots to achieve a set ``max_errors``. The final
    key will be in the form like ``obs_mistake_mask=EE_``, where `EE_` will be
    substituted by the actual observable combination.

    We estimate the probabilities of each possible observable combination by
    studying their correlation surfaces. Consider the correlation surfaces of
    some observables, an error chain crossing some fragment of correlation surface
    can flip all the observables iff. the fragment belongs to all the correlation
    surfaces, i.e. the intersection (AND) of the correlation surfaces. For each
    fragment, we find all the observables that contain it and generate the
    combination key. We then increase the frequency of the key, which means a
    single error chain at the fragment can flip all the observables. We use the
    number of fragments that can flip a observable comb as the proxy of the
    probability of the observable comb.

    Args:
        observables: The list of observables.

    Returns:
        The heuristic custom error key.

    """
    # observable with minimal correlation surface area
    minimal_area_obs = min(observables, key=lambda x: x.area)

    # Estimate the probabilities of each observable combination that contains
    # the least frequent observable
    obs_comb_to_prob: dict[str, int] = {}
    for fragment in minimal_area_obs.span:
        obs_comb: list[str] = ["E" if fragment in obs.span else "_" for obs in observables]
        key = "obs_mistake_mask=" + "".join(obs_comb)
        obs_comb_to_prob[key] = obs_comb_to_prob.get(key, 0) + 1
    # return the most frequent observable combination
    return max(obs_comb_to_prob.keys(), key=lambda x: obs_comb_to_prob[x])



from tqec.plaquette.compilation.passes.transformer import (
    InstructionCreator,
    ScheduledCircuitTransformation,
    ScheduledCircuitTransformationPass,
    ScheduleFunction,
    ScheduleOffset,
)
from tqec.plaquette.compilation.passes.transformer.simplifiers import (
    SelfInverseGateSimplification,
)
from tqec.utils.enums import Basis


class ChangeControlledGateBasisPass(ScheduledCircuitTransformationPass):
    def __init__(
        self, basis: Basis, bcsched1: ScheduleFunction, bcsched2: ScheduleFunction
    ) -> None:
        """Change ``CX`` or ``CZ`` gates to the provided basis.

        Args:
            basis: the target basis. If ``X``, all ``CZ`` gates will be changed
                to ``CX``. If ``Z``, all ``CX`` gates will be changed to ``CZ``.
            bcsched1: basis change schedule 1, a description of the schedule at
                which the potential first basis change (``H`` gate applied
                **before** the controlled gate) should be inserted.
            bcsched2: basis change schedule 2, a description of the schedule at
                which the potential second basis change (``H`` gate applied
                **after** the controlled gate) should be inserted.

        """
        ibasis = Basis.X if basis == Basis.Z else Basis.Z
        transformations = [
            ScheduledCircuitTransformation(
                f"C{ibasis.value.upper()}",
                {
                    ScheduleOffset(0): [InstructionCreator(f"C{basis.value.upper()}")],
                    bcsched1: [InstructionCreator("H", lambda trgts: trgts[1::2])],
                    bcsched2: [InstructionCreator("H", lambda trgts: trgts[1::2])],
                },
                instruction_simplifier=SelfInverseGateSimplification("H"),
            )
        ]
        super().__init__(transformations)
